
from typing import Dict, Any, Set, Optional, Union
from collections import defaultdict
from ..struct import FundScoreParam, ScoreFilter, ScorePenaltyParams
from .data_tables import FundDataTables
import dataclasses
import pandas as pd
import re
import datetime
import time
import math
import numpy as np
from copy import deepcopy

@dataclasses.dataclass
class ScoreFunc:

    alpha: float = 0  # weight of alpha
    beta: float = 0  # weight of abs(1-beta), or beta's deviation from 1
    track_err: float = 0  # weight of track_err
    fee_rate: float = 0  # weight of fee_rate
    last_year_alpha: float = 0
    last_year_beta: float = 0
    last_two_year_alpha: float = 0
    last_two_year_beta: float = 0
    down_risk: float = 0
    info_ratio: float = 0
    ret_over_period: float = 0
    treynor: float = 0
    mdd: float = 0
    m_square: float = 0
    var_: float = 0
    sharpe: float = 0
    annual_ret: float = 0
    annual_vol: float = 0
    time_ret: float = 0
    alpha_w: float = 0
    beta_w: float = 0
    track_err_w: float = 0
    beta_m: float = 0
    calmar_ratio_m: float = 0
    information_ratio_m: float = 0
    jensen_alpha_m: float = 0
    sharpe_ratio_m: float = 0
    treynor_ratio_m: float = 0

    def get(self, data):
        return  self.alpha * data.alpha \
                + self.track_err * data.track_err \
                + self.fee_rate * data.fee_rate \
                + self.beta * abs(1-data.beta) \
                + self.down_risk * data.down_risk \
                + self.info_ratio * data.info_ratio \
                + self.ret_over_period * data.ret_over_period \
                + self.treynor * data.treynor \
                + self.mdd * data.mdd \
                + self.m_square * data.m_square \
                + self.var_ * data.var_ \
                + self.sharpe * data.sharpe \
                + self.annual_ret * data.annual_ret \
                + self.annual_vol * data.annual_vol \
                + self.time_ret * data.time_ret

    def get_all(self, data):
        return  self.alpha * data.alpha \
                + self.track_err * data.track_err \
                + self.fee_rate * data.fee_rate \
                + self.down_risk * data.down_risk \
                + self.info_ratio * data.info_ratio \
                + self.ret_over_period * data.ret_over_period \
                + self.treynor * data.treynor \
                + self.mdd * data.mdd \
                + self.m_square * data.m_square \
                + self.var_ * data.var_ \
                + self.sharpe * data.sharpe \
                + self.annual_ret * data.annual_ret \
                + self.annual_vol * data.annual_vol \
                + self.time_ret * data.time_ret \
                + self.beta * abs(1-data.beta) \
                + self.alpha_w * data.alpha_w \
                + self.track_err_w * data.track_err_w \
                + self.calmar_ratio_m * data.calmar_ratio_m \
                + self.information_ratio_m * data.information_ratio_m \
                + self.jensen_alpha_m * data.jensen_alpha_m \
                + self.sharpe_ratio_m * data.sharpe_ratio_m \
                + self.treynor_ratio_m * data.treynor_ratio_m \
                + self.beta_w * abs(1-data.beta_w) \
                + self.beta_m * abs(1-data.beta_m) \
                + self.last_year_alpha * data.last_year_alpha \
                + self.last_year_beta * data.last_year_beta \
                + self.last_two_year_alpha * data.last_two_year_alpha \
                + self.last_two_year_beta * data.last_two_year_beta

class FundScoreManager:

    def __init__(self, score_filter=ScoreFilter(), score_penalty_param=ScorePenaltyParams()):
        self.params = None
        self.dts = None
        # 归一化之后的评分
        # self.funcs = {
        #     'hs300': ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1),
        #     'csi500': ScoreFunc(alpha=0.8, beta=-0.2),
        #     'gem': ScoreFunc(alpha=0.05, beta=-0.50, track_err=-0.45),
        #     'national_debt': ScoreFunc(alpha=0.32, beta=-0.33, track_err=-0.35),
        #     'mmf': ScoreFunc(alpha=0.6, fee_rate=-0.3, track_err=-0.1),
        #     'sp500rmb': ScoreFunc(fee_rate=-0.2, track_err=-0.8),
        #     'gold': ScoreFunc(alpha=0.15, fee_rate=0, track_err=-0.85),
        # }
        '''
        # weekly indicator mean of std on each day each index
        {'alpha_std': {'csi500': 0.081316,
                        'gem': 0.060555,
                        'gold': 0.014478,
                        'hs300': 0.050694,
                        'mmf': 0.0063286,
                        'national_debt': 0.0751936,
                        'sp500rmb': 0.0163182},
        'fee_rate_std': {'csi500': 0.0031425,
                            'gem': 0.0043791,
                            'gold': 0.0031,
                            'hs300': 0.002674,
                            'mmf': 0.00083814,
                            'national_debt': 0.002321,
                            'sp500rmb': 0.0022616},
        'track_err_std': {'csi500': 0.0568387,
                        'gem': 0.0554576,
                        'gold': 0.024641,
                        'hs300': 0.04155,
                        'mmf': 0.000555,
                        'national_debt': 0.052457,
                        'sp500rmb': 0.0140498}}
        
        # 日因子均值
        {'gold': {'alpha_mean': -0.00145, 'fee_mean': 0.00793, 'track_mean': 0.06711},
        'sp500rmb': {'alpha_mean': -0.0038, 'fee_mean': 0.01029,'track_mean': 0.02805},
        'mmf': {'alpha_mean': 0.00998, 'fee_mean': 0.00325, 'track_mean': 0.0021},
        'csi500': {'alpha_mean': 0.02104, 'fee_mean': 0.00806, 'track_mean': 0.07783},
        'hs300': {'alpha_mean': 0.02763, 'fee_mean': 0.00889, 'track_mean': 0.04522},
        'national_debt': {'alpha_mean': 0.03921, 'fee_mean': 0.00573, 'track_mean': 0.02394},
        'gem': {'alpha_mean': -0.02174, 'fee_mean': 0.00868, 'track_mean': 0.05726}}

        # 日因子 std
        'csi500': {'alpha_std': 0.08291, 'fee_rate': 0.0031, 'track_err': 0.06534},
        'hs300': {'alpha_std': 0.0444, 'fee_rate': 0.00266, 'track_err': 0.03939},
        'sp500rmb': {'alpha_std': 0.01625, 'fee_rate': 0.00228, 'track_err': 0.01088},
        'mmf': {'alpha_std': 0.00917, 'fee_rate': 0.00084, 'track_err': 0.00123},
        'national_debt': {'alpha_std': 0.08303,'fee_rate': 0.00229,'track_err': 0.05373},
        'gem': {'alpha_std': 0.04332, 'fee_rate': 0.00464, 'track_err': 0.04749},
        'gold': {'alpha_std': 0.03312, 'fee_rate': 0.00311, 'track_err': 0.04839}

        # 周因子std
        self.funcs_weekly = {
            'hs300': ScoreFunc(alpha=0.3/0.05069, beta=-0.6, fee_rate=-0.1/0.002674),
            'csi500': ScoreFunc(alpha=0.8/0.081316, beta=-0.2),
            'gem': ScoreFunc(alpha=0.05/0.060555, beta=-0.50, track_err=-0.45/0.0554576),
            'national_debt': ScoreFunc(alpha=0.3/0.07519359, beta=-0.25, track_err=-0.45/0.052457),
            'mmf':ScoreFunc(alpha=0.6/0.0063286, fee_rate=-0.3/0.0008381, track_err=-0.1/0.000555),
            'sp500rmb': ScoreFunc(fee_rate=-0.2/0.002261574, track_err=-0.8/0.014049778),
            'gold': ScoreFunc(alpha=0.15/0.01447791, fee_rate=-0/0.0031002, track_err=-0.85/0.02464),
        }
 
        '''
        self.funcs = {
            'hs300': ScoreFunc(alpha=0.6/ 0.0444, beta=-0.3, track_err=-0.0/0.03939 ,fee_rate=-0.0/ 0.00266),
            'csi500': ScoreFunc(alpha=0.9/ 0.08291, beta=-0.1),
            'gem':ScoreFunc(alpha=0.3/0.04332, beta=-0.50, track_err=-0.2/0.04749),
            'national_debt': ScoreFunc(alpha=0.3/0.08303, beta=-0.5, track_err=-0.4/0.05373, fee_rate=-0.0/0.00229),
            'mmf': ScoreFunc(alpha=0.3/0.00917, beta=-0.2, fee_rate=-0.1/0.00084, track_err=-0.4/0.00123),
            'sp500rmb': ScoreFunc(fee_rate=-0.2/0.00228, track_err=-0.8/0.01088),
            'gold': ScoreFunc(alpha=0.15/0.03312, fee_rate=0, track_err=-0.85/0.04839),
            'hsi': ScoreFunc(alpha=0.3, fee_rate=-0.3, track_err=-0.4),# set default score equation
        }
        self.score_cache = defaultdict(dict)
        self.score_raw_cache = defaultdict(dict)
        self.score_filter = score_filter
        self.score_penalty_param = score_penalty_param
        self._size_and_com_hold_cache: Dict[Any, pd.Index] = {}

    def set_param(self, score_param: FundScoreParam):
        self.params = score_param

    def set_dts(self, dts: FundDataTables):
        self.dts = dts
        self.fund_id_to_name = self.dts.fund_info[['fund_id', 'desc_name']].set_index('fund_id').to_dict()['desc_name']
        self.fund_name_to_id = self.dts.fund_info[['fund_id', 'desc_name']].set_index('desc_name').to_dict()['fund_id']

    def _get_fund_score(self, dt, index_id, is_filter_c, score_func: Optional[ScoreFunc] = None) -> dict:
        assert self.params and self.dts, 'cannot provide fund_score without params or data tables'
        try:
            # pivot_table会把index以外各列都为nan的行自动干掉(感觉应该是bug，而且dropna=False的行为也不合理)
            # 因此这里如果找不到的话直接返回空
            cur_d = self.dts.index_fund_indicator_pack.loc[index_id, dt]
        except KeyError:
            return {}, {}
        return self._get_score(index_id=index_id, dt=dt, cur_d=cur_d, is_filter_c=is_filter_c, score_func=score_func)

    def get_fund_score(self, dt, index_id, is_filter_c, score_func: Optional[Union[ScoreFunc, str]] = None):
        if score_func is not None and score_func:
            # 指定了score_func，缓存里肯定没有，直接算一遍然后返回
            return self._get_fund_score(dt, index_id, is_filter_c, score_func)
        try:
            # 先尝试从缓存里取
            return self.score_cache[dt][index_id], self.score_raw_cache[dt][index_id]
        except KeyError:
            # 获取不到的话则算一遍，注意这里不存缓存
            return self._get_fund_score(dt, index_id, is_filter_c)

    def get_fund_scores(self, dt, index_list, is_filter_c=True, fund_score_funcs: Optional[Dict[str, Union[ScoreFunc, str]]] = None) -> dict:
        index_score = self.score_cache[dt]
        index_score_raw = self.score_raw_cache[dt]
        if not index_score or not index_score_raw:
            # 先用默认的fund_score_funcs对所有的index算一下
            for index_id in index_list:
                # 这里由于index_score和index_score_raw都是dict, 所以我们这样赋值可以更新到self.score_cache和self.score_raw_cache
                index_score[index_id], index_score_raw[index_id] = self._get_fund_score(dt, index_id, is_filter_c)

        # 没有指定fund_score_funcs，可以直接返回
        if fund_score_funcs is None or not fund_score_funcs:
            return index_score, index_score_raw

        # 如果指定了fund_score_funcs，需要重新算一下其中指定的index
        score = {}
        score_raw = {}
        for index_id, func in fund_score_funcs.items():
            score[index_id], score_raw[index_id] = self._get_fund_score(dt, index_id, is_filter_c, func)
        # 将指定的index的和默认的index组合到一起返回
        # python3.9可以直接用"|"运算符
        score.update({index_id: v for index_id, v in index_score.items() if index_id not in score})
        score_raw.update({index_id: v for index_id, v in index_score_raw.items() if index_id not in score_raw})
        return score, score_raw

    def _wrapper(self, df: pd.DataFrame, is_filter_c: bool, use_weekly_monthly_indicators: bool):
        # T日大类下有很多基金，这里取一下它们的datetime和index_id，因为值都是一样的所以取array[0]就OK
        dt = df.datetime.array[0]
        index_id: str = df.index_id.array[0]
        # 这里把index换成了fund_id
        score, score_raw = self._get_score(index_id=index_id, dt=dt, cur_d=df.set_index('fund_id').drop(columns=['datetime', 'index_id']), is_filter_c=is_filter_c, use_weekly_monthly_indicators=use_weekly_monthly_indicators)
        self.score_cache[dt][index_id] = score
        self.score_raw_cache[dt][index_id] = score_raw

    def pre_calculate(self, is_filter_c=True, use_weekly_monthly_indicators=False):
        # self.dts.fund_indicator.pivot_table(index=['datetime', 'index_id'])
        # 上面这一行会耗时10秒左右，且之后遍历index(datetime-index_id pair)也会比下边groupby慢一些
        # 这里应该没必要先groupby datetime，再groupby index_id（如果真这样可能更慢吧）
        self.dts.fund_indicator.groupby(by=['datetime', 'index_id']).apply(lambda x: self._wrapper(x, is_filter_c, use_weekly_monthly_indicators))
        
    def _get_score(self, index_id, dt, cur_d, is_filter_c, use_weekly_monthly_indicators=False, score_func: Optional[Union[ScoreFunc, str]] = None):
        fund_ids = self._filter_size_and_com_hold(dt, index_id, cur_d.index)
        if fund_ids is not cur_d.index:
            # 这里没有dropna总计可以快24秒
            cur_d = cur_d.reindex(fund_ids)
        if cur_d.shape[0] > 1:
            # 不做归一化
            func = self.funcs.get(index_id) if score_func is None else score_func
            if isinstance(func, str):
                try:
                    FundScoreManager.test_func(func)
                except Exception as e:
                    # func str不合法，直接返回空
                    print(f'eval fund score func str failed for {index_id} (err_msg){e}')
                    return {}, {}
                score_raw = cur_d.apply(FundScoreManager.get_func(func), axis=1)
            else:
                # 默认均走这个分支
                if use_weekly_monthly_indicators:
                    score_raw = cur_d.apply(lambda x: func.get_all(x), axis=1)
                else:
                    score_raw = cur_d.apply(lambda x: func.get(x), axis=1)

            if index_id in self.score_penalty_param.FilterYearIndex:
                punish_funds = cur_d[cur_d.year_length < self.score_penalty_param.JudgeYearLength].index.tolist()
                for _fund_id in punish_funds:
                    score_raw[_fund_id] += self.score_penalty_param.Penalty
                # 以下是vectorize calc, 但看起来没有明显的加速效果
                # punish_funds = pd.Series(self.score_penalty_param.Penalty, index=cur_d[cur_d.year_length < self.score_penalty_param.JudgeYearLength].index)
                # score_raw = score_raw.add(punish_funds, fill_value=0)
            score = (score_raw - score_raw.min()) / (score_raw.max() - score_raw.min())
            if is_filter_c:
                score = self._filter_score(score)
            score = score.to_dict()
            score_raw = score_raw.to_dict()
        else:
            score_raw = {cur_d.index[0]: 1}
            score = score_raw
        return score, score_raw

    def _filter_size_and_com_hold(self, dt, index_id: str, fund_ids: pd.Index):
        if index_id not in self.score_filter.FilterIndexId:
            return fund_ids
        try:
            # retrieve from cache
            _select_funds = self._size_and_com_hold_cache[dt]
        except KeyError:
            try:
                fund_com_hold_dt = self.dts.fund_com_hold.loc[dt]
            except KeyError:
                return fund_ids

            hold_select_fund_id = fund_com_hold_dt[fund_com_hold_dt < self.score_filter.CompanyHoldLimit].index
            fund_size_dt = self.dts.fund_size.loc[dt]
            size_select_fund_id = fund_size_dt[(fund_size_dt >= self.score_filter.SizeBottom)
                                               & (fund_size_dt <= self.score_filter.SizeTop)].index
            _select_funds = hold_select_fund_id.intersection(size_select_fund_id)
            # 这里做了一个缓存：T日备选的基金列表；因为多个大类可能多次取同一日的备选基金列表
            # 这个缓存没有清理，但看起来应该没问题
            self._size_and_com_hold_cache[dt] = _select_funds
        select_funds = _select_funds.intersection(fund_ids)
        if select_funds.empty:
            return fund_ids
        else:
            return select_funds

    def _filter_score(self, score_series: pd.Series) -> pd.Series:
        # 同一天同资产 同基金A B 在 选B
        # 同一天同资产 同基金A C 在 选C
        # 这里比之前总计快6秒左右
        index_fund_name: Set[str] = {self.fund_id_to_name[_] for _ in score_series.index}
        for fund_name in index_fund_name:
            if 'B' in fund_name or 'C' in fund_name:
                prefered_fund, count = re.subn(r'[BC]$', 'A', fund_name)
                if count != 0 and prefered_fund in index_fund_name:
                    score_series[self.fund_name_to_id[prefered_fund]] = 0
        return score_series

    def _white_and_black_list_filter(self, score, score_raw, disproved_set):
        score = self._white_black_func(deepcopy(score), disproved_set)
        score_raw = self._white_black_func(deepcopy(score_raw), disproved_set)
        return score, score_raw

    def _white_black_func(self, score, disproved_set):
        del_list = []
        for index_id in score:
            for fund_id in score[index_id]:
                if fund_id in disproved_set:
                    # 之前是给非白名单 黑名单的基金给分数惩罚，但是基金数量不足时，会交易
                    # 现在改为强逻辑，直接删掉
                    #score[index_id][fund_id] += self.score_penalty_param.BlackListPenalty
                    del_list.append([index_id, fund_id])
        for del_i in del_list:
            del score[del_i[0]][del_i[1]]
        return score

    @staticmethod
    def test_func(func_str):
        for item in ScoreFunc.__dataclass_fields__.keys():
            locals()[item] = 0
        # if we cannot calculate, it just breaks
        eval(func_str)
        return True

    @staticmethod
    def get_func(func_str):
        new_func = func_str
        for item in ScoreFunc.__dataclass_fields__.keys():
            new_func = new_func.replace(item, 'x.' + item)
        return lambda x: eval(new_func)

    @staticmethod
    def score_calc(func_str, dm, index_id, dt):
        cur_d = dm.dts.index_fund_indicator_pack.loc[index_id, dt]
        f = FundScoreManager.get_func(func_str)
        return cur_d.apply(f, axis=1)



#!/usr/bin/env python
# coding: utf-8

import numpy as np
import pandas as pd
import rqdatac as rq
import pickle as pkl
from typing import List, Set, Tuple, Dict
from operator import itemgetter
from collections import defaultdict
from .EMQuantAPI_Python.python3.EmQuantAPI import *
import datetime
import json
import os
import csv
import traceback
from sqlalchemy import distinct
from sqlalchemy.orm import sessionmaker
from ...wrapper.mysql import RawDatabaseConnector
from ...view.raw_models import *
from ...api.basic import BasicDataApi
from ...api.raw import RawDataApi
from .raw_data_helper import RawDataHelper
from ....constant import SectorType, IndexPriceSource, CodeFeeMode, INDEX_PRICE_EXTRA_TRADE_DAYS


class EmRawDataDownloader:

    _MACROECONOMIC_MAP = {
        'EMM00166216': 'RA0000.IR',
        'EMM00166217': 'RA0001.IR',
        'EMM00166218': 'RA0002.IR',
        'EMM00166219': 'RA0003.IR',
        'EMM00166220': 'RA0004.IR',
        'EMM00166221': 'RA0005.IR',
        'EMM00166222': 'RA0006.IR',
        'EMM00166228': 'RA0007.IR',
    }
    _MACROECONOMIC_MAP_I = {v: k for k, v in _MACROECONOMIC_MAP.items()}

    def __init__(self, data_helper):
        self._data_helper = data_helper
        # ForceLogin
        # 取值0，当线上已存在该账户时，不强制登录
        # 取值1，当线上已存在该账户时，强制登录，将前一位在线用户踢下线
        options = 'ForceLogin=1'
        loginResult = c.start(options, mainCallBack=self._main_callback)
        if(loginResult.ErrorCode != 0):
            print('EM login failed')
            exit()

        self._scale_change_dates = ['0331', '0630', '0930', '1231']

    def __del__(self):
        c.stop()

    def _main_callback(self, quantdata):
        '''
        mainCallback 是主回调函数，可捕捉连接错误
        该函数只有一个为c.EmQuantData类型的参数quantdata
        :param quantdata: c.EmQuantData
        '''
        print(f'MainCallback: {quantdata}')

    def _get_date_list(self, start_date, end_date):
        start_datetime = datetime.datetime.strptime(start_date, '%Y%m%d')
        end_datetime = datetime.datetime.strptime(end_date, '%Y%m%d')
        date_list = []
        while start_datetime <= end_datetime:
            date_list.append(start_datetime.strftime('%Y%m%d'))
            start_datetime += datetime.timedelta(days=1)
        return date_list

    def em_fund_nav_history(self, nav_file_dir):
        try:
            index = 1
            for filename in os.listdir(nav_file_dir):
                if filename.endswith(".xls"):
                    print(f'{index}: {filename}')
                    index += 1
                    fund_nav = pd.concat(pd.read_excel(os.path.join(nav_file_dir, filename), sheet_name=None),
                                        ignore_index=True)
                    fund_nav = fund_nav.drop(['简称'], axis=1)
                    fund_nav = fund_nav.rename(columns={
                        '代码': 'CODES',
                        '时间': 'DATES',
                        '单位净值(元)': 'ORIGINALUNIT',
                        '累计净值(元)': 'ORIGINALNAVACCUM',
                        '复权净值(元)': 'ADJUSTEDNAV',
                        '万份基金单位收益(元)': 'UNITYIELD10K',
                        '7日年化收益率(%)': 'YIELDOF7DAYS'
                    })

                    fund_nav = fund_nav[(fund_nav['DATES'] != '--')
                                        & (~fund_nav['DATES'].isnull())]
                    fund_nav = fund_nav.replace('--', np.nan)
                    self._data_helper._upload_raw(fund_nav, EmFundNav.__table__.name)
                else:
                    continue
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index_price_history(self, em_ids, start_date, end_date):
        em_id_str = ','.join(em_ids) if isinstance(em_ids, list) else em_ids
        df = c.csd(em_id_str, "CLOSE", start_date, end_date,
                   "period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1")
        if not isinstance(df, pd.DataFrame):
            if df.ErrorCode != 0:
                print(f'failed to get index price history: {df.ErrorMsg} (index_ids){em_ids}')
            return

        df = df.reset_index().rename(columns={
            'CODES': 'em_id',
            'DATES': 'datetime',
            'CLOSE': 'close'
        })
        df = df[df['close'].notna()]
        self._data_helper._upload_raw(df, EmIndexPrice.__table__.name)

    def em_fund_scale_history(self):
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                dates = db_session.query(
                    distinct(EmFundNav.DATES)
                ).all()

                dates = [d for d, in dates]
                dates.sort()
                start_date = dates[0]
                for i in range(1, 100):
                    dates.insert(0, start_date - datetime.timedelta(days=i))

                for d in dates:
                    month_day = d.strftime('%m%d')
                    if month_day not in self._scale_change_dates:
                        continue

                    print(d)

                    fund_ids = db_session.query(
                        distinct(EmFundNav.CODES)
                    ).filter(
                        EmFundNav.DATES >= d,
                        EmFundNav.DATES <= d + datetime.timedelta(days=100)
                    ).all()

                    fund_ids = [f for f, in fund_ids]
                    fund_id_str = ','.join(fund_ids)

                    df = c.css(fund_id_str, "FUNDSCALE",
                               f"EndDate={d},Ispandas=1").reset_index()
                    df['DATES'] = d

                    self._data_helper._upload_raw(df, EmFundScale.__table__.name)
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))
                return None

    def em_fund_holding_rate_history(self):
        # 机构持有比例半年更新，到期日未必有数据，会有延迟， 含已终止基金
        fund_id_result = c.sector('507010', '2020-06-10')
        fund_list = fund_id_result.Codes
        fund_id_str = ','.join(fund_list)
        y_list = [str(y) for y in range(2010, 2020,1)]
        d_list = []
        for y in y_list:
            for d in ['-06-30', '-12-31']:
                d_list.append(y+d)
        for d in d_list:
            data=c.css(fund_id_str,"HOLDPERSONALHOLDINGPCT,HOLDINSTIHOLDINGPCT",f"ReportDate={d} ,Ispandas=1")
            data['DATES'] = d
            data = data.reset_index().dropna()
            self._data_helper._upload_raw(data, EmFundHoldingRate.__table__.name)
            print(f'finish {d} ')

    def em_fund_nav(self, start_date, end_date, fund_id_list: Tuple[str] = ()):
        try:
            # TODO: 应该使用每一天当天的成分列表, 可能所有的sector都需要这样
            # Get all fund ids, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=5)
            start_date = start_date.strftime('%Y%m%d')
            trade_date_df = RawDataApi().get_em_tradedates(start_date, end_date)
            if trade_date_df is None:
                print('failed to get em trade dates for em_fund_nav')
                return
            if not fund_id_list:
                fund_id_result = c.sector('507013', end_date)
                if fund_id_result.ErrorCode != 0:
                    print(f'Failed to get fund id list: {fund_id_result.ErrorMsg}')
                    return
                fund_id_list = fund_id_result.Codes
            fund_id_str = ','.join(fund_id_list)
            df_list: List[pd.DataFrame] = []
            for date in trade_date_df.TRADEDATES:
                # 基金 万份基金单位收益 7日年化收益率
                data1 = c.css(fund_id_str, "10KUNITYIELD,YIELDOF7DAYS", f"EndDate={date},Ispandas=1")
                if not isinstance(data1, pd.DataFrame):
                    if data1.ErrorCode != 0:
                        print(f'Failed to get fund nav data1: {data1.ErrorMsg}')
                    return
                # 基金 单位净值(原始披露) 累计单位净值(原始披露) 复权单位净值
                data2 = c.css(fund_id_str, "ORIGINALUNIT,ORIGINALNAVACCUM,NAVADJ", f"TradeDate={date},Ispandas=1")
                if not isinstance(data2, pd.DataFrame):
                    if data2.ErrorCode != 0:
                        print(f'Failed to get fund nava data2: {data2.ErrorMsg}')
                    return
                df = data2.drop(columns='DATES').join(data1.drop(columns='DATES'), how='outer')
                df = df[df[['ORIGINALUNIT', 'ORIGINALNAVACCUM', 'NAVADJ', '10KUNITYIELD', 'YIELDOF7DAYS']].notna().any(axis=1)]
                df = df.rename(columns={'10KUNITYIELD': 'UNITYIELD10K', 'NAVADJ': 'ADJUSTEDNAV'}).reset_index()
                df['DATES'] = pd.to_datetime(date).date().strftime('%Y-%m-%d')
                df_list.append(df)

            df_total = pd.concat(df_list)
            # 这里删掉之前几天的数据，然后再重刷一遍
            # TODO: 最好可以原子提交下边两步操作
            self.delete_em_fund_nav(start_date, end_date)
            self._data_helper._upload_raw(df_total, EmFundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def delete_em_fund_nav(self, start_date, end_date):
        try:
            RawDataApi().delete_em_fund_nav(start_date, end_date)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_fund_status(self, end_date: str):
        try:
            fund_list = RawDataApi().get_em_fund_list(end_date, limit=1)
            if fund_list is None:
                print(f'failed to get fund list for fund status')
                return False
            fund_list = fund_list.all_live_fund_list.array[0]
            data1 = c.css(fund_list, 'FUNDSCALE', f'EndDate={end_date},Ispandas=1')
            if not isinstance(data1, pd.DataFrame):
                if data1.ErrorCode != 0:
                    print(f'failed to get fund status data1: {data1.ErrorMsg}')
                return False

            data2 = c.css(fund_list, 'PURCHSTATUS,REDEMSTATUS,LPLIMIT,OTCLPLIMIT,TRADESTATUS', f'TradeDate={end_date},Ispandas=1')
            if not isinstance(data2, pd.DataFrame):
                if data2.ErrorCode != 0:
                    print(f'failed to get fund status datta2: {data2.ErrorMsg}')
                return False

            df = data1.drop(columns='DATES').join(data2.drop(columns='DATES'), how='outer')
            df = df[df.notna().any(axis=1)]
            if df.empty:
                return True

            df = df.reset_index()
            df['DATES'] = end_date
            df = df.astype({'FUNDSCALE': 'float64', 'LPLIMIT': 'float64', 'OTCLPLIMIT': 'float64'})
            # 获取当前的fund status
            now_df = RawDataApi().get_em_fund_status(end_date=end_date)
            if now_df is not None:
                # 过滤出来新增基金status以及旧基金的status发生变化的情况
                df = df.merge(now_df.drop(columns=['_update_time']), how='left', on=['CODES', 'FUNDSCALE', 'LPLIMIT', 'OTCLPLIMIT', 'PURCHSTATUS', 'REDEMSTATUS', 'TRADESTATUS'], indicator=True, validate='one_to_many')
                df = df[df._merge == 'left_only'].drop(columns=['_merge', 'DATES_y']).rename(columns={'DATES_x': 'DATES'})
            self._data_helper._upload_raw(df, EmFundScale.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index_price(self, em_ids: Set[str], start_date, end_date):
        try:
            # 每次重刷15个交易日的数据
            data = c.getdate(start_date, -INDEX_PRICE_EXTRA_TRADE_DAYS)
            if data.ErrorCode != 0:
                print(f'[em_index_price] failed to get trade dates: {data.ErrorMsg} (start_date){start_date}')
                return False
            start_date = pd.to_datetime(data.Data[0]).strftime('%Y%m%d')

            # 不支持跨品种的index同时请求，所以需要拆出来分开去做
            df_list = []
            # 利率部分
            ir_index = []
            for one in ('LIGBP1M.IR', 'HIBOR3M.IR'):
                try:
                    em_ids.remove(one)
                except KeyError:
                    continue
                ir_index.append(one)
            if ir_index:
                data = c.csd(','.join(ir_index), "CLOSE", start_date, end_date, "period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")
                if not isinstance(data, pd.DataFrame):
                    if data.ErrorCode != 0:
                        print(f'[em_index_price] failed to get index price: {data.ErrorMsg} (index_id){ir_index}')
                    return False
                df_list.append(data)

            # 期货部分
            future_index = []
            for one in ('AU9999.SGE', 'AU0.SHF'):
                try:
                    em_ids.remove(one)
                except KeyError:
                    continue
                future_index.append(one)
            if future_index:
                data = c.csd(','.join(future_index), "CLOSE", start_date, end_date, "period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")
                if not isinstance(data, pd.DataFrame):
                    if data.ErrorCode != 0:
                        print(f'[em_index_price] failed to get index price: {data.ErrorMsg} (index_id){future_index}')
                    return False
                df_list.append(data)

            # 常规指数部分
            data = c.csd(','.join(em_ids), "CLOSE", start_date, end_date, "period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'[em_index_price] failed to get index price: {data.ErrorMsg} (index_id){em_ids}')
                return False

            df = pd.concat(df_list + [data]).reset_index().rename(columns={
                'CODES': 'em_id',
                'DATES': 'datetime',
                'CLOSE': 'close'
            })
            df = df[df['close'].notna()]
            # TODO: 最好可以原子提交下边两步操作
            RawDataApi().delete_em_index_price(list(em_ids)+ir_index+future_index, start_date, end_date)
            self._data_helper._upload_raw(df, EmIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index_info(self, index_ids: List[str]):
        try:
            data = c.css(','.join(index_ids), 'SHORTNAME,NAME,PUBLISHDATE,MAKERNAME,INDEXPROFILE', 'Ispandas=1')
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'failed to get index info: {data.ErrorMsg}')
                return False

            df = data.drop(columns='DATES').reset_index().rename(columns=lambda x: EmIndexInfo.__getattribute__(EmIndexInfo, x).name)
            # 统一设置一个默认值，之后可以按需手工修改
            df['price_source'] = IndexPriceSource.default.name
            # 更新到db
            self._data_helper._upload_raw(df, EmIndexInfo.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_macroeconomic_price(self, start_date, end_date):
        try:
            index_info_df = RawDataApi().get_em_index_info()
            if index_info_df is None:
                print('failed to get index info for macroeconomic price')
                return False
            index_info_df = index_info_df[index_info_df.price_source == IndexPriceSource.macroeconomic]
            macroeconomic_list = [self._MACROECONOMIC_MAP_I[one] for one in index_info_df.em_id.to_list() if one in self._MACROECONOMIC_MAP_I]
            data = c.edb(','.join(macroeconomic_list), f'StartDate={start_date},EndDate={end_date},IsLatest=0,RowIndex=2,Ispandas=1')
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    if data.ErrorCode == 10000009:
                        print(f'no macroeconomic price data during {start_date} and {end_date}')
                        return True
                    print(f'failed to get macroeconomic price: {data.ErrorMsg}')
                return False

            df = data.reset_index().rename(columns={
                'CODES': 'em_id',
                'DATES': 'datetime',
                'RESULT': 'close'
            })
            df = df[df.close.notna()]
            df['em_id'] = df['em_id'].map(self._MACROECONOMIC_MAP)
            # 更新到db
            self._data_helper._upload_raw(df, EmIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index(self, start_date, end_date):
        # ----------------------
        # H11073.CSI,信用债
        # SPX.GI,标普500
        # GDAXI.GI,德国DAX
        # N225.GI,日经225
        # 000012.SH,利率债
        # 000300.SH,沪深300
        # 000905.SH,中证500
        # 399006.SZ,创业板
        # 801181.SWI,房地产开发申万
        # ----------------------
        # USDCNY.IB,美元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # JPYCNY.IB,日元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # EURCNY.IB,欧元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # ----------------------
        # AU0.SHF,沪金主力连续,TODO: always get None
        # BC.ICE,布伦特原油当月连续,TODO: always get None
        # ----------------------
        codes = (
                 'H11073.CSI,SPX.GI,GDAXI.GI,N225.GI,000012.SH,000300.SH,000905.SH,399006.SZ,801181.SWI,'
                 'USDCNY.IB,JPYCNY.IB,EURCNY.IB,'
                #  'AU0.SHF,BC.ICE'
        )
        df = c.csd(codes, 'CLOSE', start_date, end_date, 'Ispandas=1')
        if df is not None:
            df.reset_index(inplace=True)
            print(df)
            # self._data_helper._upload_raw(df, EmIndexPrice.__table__.name)

    # 从em客户端导出数据后读取并处理
    def em_stock_price_history(self, stock_price_file_dir: str):
        for index, filename in enumerate(os.listdir(stock_price_file_dir)):
            if not filename.endswith(".xls"):
                continue
            index += 1
            print(f'{index}: {filename}')
            stock_price = pd.concat(pd.read_excel(os.path.join(stock_price_file_dir, filename), sheet_name=None, na_values='--'),
                                    ignore_index=True)
            # 将列名字改为从API获取到数据时的列名字，便于后边统一处理
            stock_price = stock_price.drop(['简称'], axis=1).rename(columns={
                '代码': 'CODES',
                '时间': 'DATES',
                '开盘价(元)': 'OPEN',
                '收盘价(元)': 'CLOSE',
                '最高价(元)': 'HIGH',
                '最低价(元)': 'LOW',
                '前收盘价(元)': 'PRECLOSE',
                '均价(元)': 'AVERAGE',
                '成交量(股)': 'VOLUME',
                '成交金额(元)': 'AMOUNT',
                '换手率(%)': 'TURN',
                '交易状态': 'TRADESTATUS',
                '内盘成交量': 'BUYVOL',
                '外盘成交量': 'SELLVOL',
            })

            # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
            stock_price = stock_price.loc[stock_price['TRADESTATUS'] != '未上市'].loc[~stock_price['DATES'].isna()].rename(columns=lambda x: EmStockPrice.__getattribute__(EmStockPrice, x).name)
            # 更新到db
            self._data_helper._upload_raw(stock_price, EmStockPrice.__table__.name)

    def em_stock_price(self, start_date: str, end_date: str, stock_list: Tuple[str] = ()):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_price] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_price] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            if not stock_list:
                # 获取所有A股股票的ID, 功能函数-板块成分
                # http://quantapi.eastmoney.com/Cmd/Sector?from=web
                # 全部A股股票
                stock_id_result = c.sector('001004', end_date)
                if stock_id_result.ErrorCode != 0:
                    print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                    return
                stock_list = stock_id_result.Codes

            df_list = []
            for date in tradedates.Data:
                # 获取个股股价信息
                temp_df = c.csd(stock_list,
                                'OPEN,CLOSE,HIGH,LOW,PRECLOSE,AVERAGE,VOLUME,AMOUNT,TURN,TRADESTATUS,TNUM,BUYVOL,SELLVOL',
                                date, date, 'period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock price info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
                df_list.append(temp_df.loc[temp_df['TRADESTATUS'] != '未上市'].loc[~temp_df['DATES'].isna()])

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockPrice.__getattribute__(EmStockPrice, x).name)
            # 更新到db
            self._data_helper._upload_raw(df, EmStockPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    # 从em客户端导出数据后读取并处理
    def em_stock_post_price_history(self, stock_post_price_file_dir: str):
        for index, filename in enumerate(os.listdir(stock_post_price_file_dir)):
            if not filename.endswith(".xls"):
                continue
            index += 1
            print(f'{index}: {filename}')
            stock_price = pd.concat(pd.read_excel(os.path.join(stock_post_price_file_dir, filename), sheet_name=None, na_values='--'),
                                    ignore_index=True)
            # 将列名字改为从API获取到数据时的列名字，便于后边统一处理
            stock_price = stock_price.drop(['简称'], axis=1).rename(columns={
                '代码': 'CODES',
                '时间': 'DATES',
                '开盘价(元)': 'OPEN',
                '收盘价(元)': 'CLOSE',
                '最高价(元)': 'HIGH',
                '最低价(元)': 'LOW',
                '前收盘价(元)': 'PRECLOSE',
                '均价(元)': 'AVERAGE',
                '成交量(股)': 'VOLUME',
                '成交金额(元)': 'AMOUNT',
                '交易状态': 'TRADESTATUS',
            })
            # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
            stock_price = stock_price.loc[stock_price['TRADESTATUS'] != '未上市'].loc[~stock_price['DATES'].isna()].rename(columns=lambda x: EmStockPostPrice.__getattribute__(EmStockPostPrice, x).name)
            # 更新到db
            self._data_helper._upload_raw(stock_price, EmStockPostPrice.__table__.name)

    def em_stock_post_price(self, start_date: str, end_date: str, stock_list: Tuple[str] = ()):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_post_price] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_post_price] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            if not stock_list:
                # 获取所有A股股票的ID, 功能函数-板块成分
                # http://quantapi.eastmoney.com/Cmd/Sector?from=web
                # 全部A股股票
                stock_id_result = c.sector('001004', end_date)
                if stock_id_result.ErrorCode != 0:
                    print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                    return
                stock_list = stock_id_result.Codes

            df_list = []
            for date in tradedates.Data:
                # 获取个股股价信息
                temp_df = c.csd(stock_list,
                                'OPEN,CLOSE,HIGH,LOW,PRECLOSE,AVERAGE,VOLUME,AMOUNT,TRADESTATUS,TAFACTOR',
                                date, date, 'period=1,adjustflag=2,curtype=1,order=1,market=0,Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock price info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
                df_list.append(temp_df.loc[temp_df['TRADESTATUS'] != '未上市'].loc[~temp_df['DATES'].isna()])

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockPostPrice.__getattribute__(EmStockPostPrice, x).name)
            # 更新到db
            self._data_helper._upload_raw(df, EmStockPostPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_stock_info(self, end_date: str):
        try:
            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return False

            # 去掉我们已经有的股票ID
            stock_info_df = RawDataApi().get_em_stock_info()
            new_stock_code = list(set(stock_id_result.Codes).difference(set(stock_info_df.stock_id.to_list())))
            if not new_stock_code:
                # 直接返回成功
                return True

            df_list: List[pd.DataFrame] = []
            # 获取无参数指标
            df_no_param = c.css(new_stock_code,
                                'NAME,LISTDATE,FINPURCHORNOT,FINSELLORNOT,STOHSTOCKCONNECTEDORNOT,SHENGUTONGTARGET,ENGNAME,COMPNAME,COMPNAMEENG',
                                'Ispandas=1')
            if not isinstance(df_no_param, pd.DataFrame):
                if df_no_param.ErrorCode != 0:
                    print(f'failed to get stock info(no param): {df_no_param.ErrorMsg}')
                return False
            # 将其中'是'/'否'的结果转换为True/False
            columns = ['FINPURCHORNOT', 'FINSELLORNOT', 'STOHSTOCKCONNECTEDORNOT', 'SHENGUTONGTARGET']
            df_no_param.loc[:, columns] = df_no_param.loc[:, columns].apply(lambda x: x.apply(lambda x: True if x == '是' else False))
            df_list.append(df_no_param)

            # 获取行业分类共有三级的指标
            df_lv4 = c.css(new_stock_code, "BLEMINDCODE,BLSWSINDCODE,SW2014CODE,EMINDCODE2016,CITICCODE2020", "ClassiFication=4,Ispandas=1")
            if not isinstance(df_lv4, pd.DataFrame):
                if df_lv4.ErrorCode != 0:
                    print(f'failed to get stock info(lv4): {df_lv4.ErrorMsg}')
                return False
            df_list.append(df_lv4)

            # 获取行业分类共有二级的指标
            df_lv3 = c.css(new_stock_code, "BLCSRCINDCODE,CSRCCODENEW", "ClassiFication=3,Ispandas=1")
            if not isinstance(df_lv3, pd.DataFrame):
                if df_lv3.ErrorCode != 0:
                    print(f'failed to get stock info(lv3): {df_lv3.ErrorMsg}')
                return False
            df_list.append(df_lv3)

            # 获取行业分类共有四级的指标
            df_lv5 = c.css(new_stock_code, "CSINDCODE2016,GICSCODE", "ClassiFication=5,Ispandas=1")
            if not isinstance(df_lv5, pd.DataFrame):
                if df_lv5.ErrorCode != 0:
                    print(f'failed to get stock info(lv5): {df_lv5.ErrorMsg}')
                return False
            df_list.append(df_lv5)

            if df_list:
                # 将每部分按列粘到一起
                df = pd.concat(df_list, axis=1)
                # 不需要存日期的列，去除索引并做列名转换
                df = df.drop(columns='DATES').reset_index().rename(columns=lambda x: EmStockInfo.__getattribute__(EmStockInfo, x).name)
                # 更新到db
                self._data_helper._upload_raw(df, EmStockInfo.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def add_columns_to_fin_fac(self, end_date: str):
        # 获取所有A股股票的ID, 功能函数-板块成分
        # http://quantapi.eastmoney.com/Cmd/Sector?from=web
        # 全部A股股票
        stock_id_result = c.sector('001004', end_date)
        if stock_id_result.ErrorCode != 0:
            print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
            return

        # 获取主营收入构成(按产品)
        df = c.css(stock_id_result.Codes, "BALANCESTATEMENT_195", f'ReportDate={end_date},type=1,Ispandas=1')
        if not isinstance(df, pd.DataFrame) and df.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df.ErrorMsg}')
            return

        count = 0
        Session = sessionmaker(RawDatabaseConnector().get_engine())
        db_session = Session()
        for row in db_session.query(EmStockFinFac):
            if row.DATES != datetime.datetime.strptime(end_date, '%Y%m%d').date():
                continue
            if row.CODES in df.index:
                row.BALANCESTATEMENT_195 = df.at[row.CODES, 'BALANCESTATEMENT_195']
                count += 1
        db_session.commit()
        db_session.close()
        print(f'append {count} value(s) to fin fac table')

    def em_stock_fin_fac(self, end_date: str):
        # 获取所有A股股票的ID, 功能函数-板块成分
        # http://quantapi.eastmoney.com/Cmd/Sector?from=web
        # 全部A股股票
        stock_id_result = c.sector('001004', end_date)
        if stock_id_result.ErrorCode != 0:
            print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
            return

        df_list: List[pd.DataFrame] = []
        # 获取无参数指标
        df_no_param = c.css(stock_id_result.Codes,
                            'EBIT,EBITDA,EXTRAORDINARY,LOWERDIANDNI,GROSSMARGIN,OPERATEINCOME,INVESTINCOME,EBITDRIVE,TOTALCAPITAL,WORKINGCAPITAL,\
                             NETWORKINGCAPITAL,TANGIBLEASSET,RETAINED,INTERESTLIBILITY,NETLIBILITY,EXINTERESTCL,EXINTERESTNCL,FCFF,FCFE,DA,FCFFDRIVE,\
                             PERFORMANCEEXPRESSPARENTNI,MBSALESCONS,GPMARGIN,NPMARGIN,EXPENSETOOR,INVTURNRATIO,ARTURNRATIO,ROEAVG,ROEWA,EPSBASIC,EPSDILUTED,\
                             BPS,BALANCESTATEMENT_25,BALANCESTATEMENT_46,BALANCESTATEMENT_93,BALANCESTATEMENT_103,BALANCESTATEMENT_141,BALANCESTATEMENT_195,\
                             BALANCESTATEMENT_140,INCOMESTATEMENT_9,INCOMESTATEMENT_48,INCOMESTATEMENT_60,INCOMESTATEMENT_61,INCOMESTATEMENT_83,INCOMESTATEMENT_85,INCOMESTATEMENT_127,\
                             INCOMESTATEMENT_14,CASHFLOWSTATEMENT_39,CASHFLOWSTATEMENT_59,CASHFLOWSTATEMENT_77,CASHFLOWSTATEMENT_82,CASHFLOWSTATEMENT_86',
                            f'ReportDate={end_date},type=1,Ispandas=1')
        if not isinstance(df_no_param, pd.DataFrame) and df_no_param.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_no_param.ErrorMsg}')
        else:
            df_list.append(df_no_param)

        # 获取主营收入构成(按产品)
        df_mb_sales_cons = c.css(stock_id_result.Codes, "MBSALESCONS", f'ReportDate={end_date},type=2,Ispandas=1')
        if not isinstance(df_mb_sales_cons, pd.DataFrame) and df_mb_sales_cons.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_mb_sales_cons.ErrorMsg}')
        else:
            df_list.append(df_mb_sales_cons.rename(columns={'MBSALESCONS': 'MBSALESCONS_P'}))

        # 获取归属于上市公司股东的扣除非经常性损益后的净利润(调整前)
        df_with_param_1 = c.css(stock_id_result.Codes, "DEDUCTEDINCOME", f'ReportDate={end_date},DataAdjustType=1,Ispandas=1')
        if not isinstance(df_with_param_1, pd.DataFrame) and df_with_param_1.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_with_param_1.ErrorMsg}')
        else:
            df_list.append(df_with_param_1.rename(columns={'DEDUCTEDINCOME': 'DEDUCTEDINCOME_BA'}))

        # 获取归属于上市公司股东的扣除非经常性损益后的净利润(调整后)
        df_with_param_2 = c.css(stock_id_result.Codes, "DEDUCTEDINCOME", f'ReportDate={end_date},DataAdjustType=2,Ispandas=1')
        if not isinstance(df_with_param_2, pd.DataFrame) and df_with_param_2.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_with_param_2.ErrorMsg}')
        else:
            df_list.append(df_with_param_2.rename(columns={'DEDUCTEDINCOME': 'DEDUCTEDINCOME_AA'}))

        # 获取TTM指标
        df_ttm = c.css(stock_id_result.Codes, 'GRTTMR,GCTTMR,ORTTMR,OCTTMR,EXPENSETTMR,GROSSMARGINTTMR,OPERATEEXPENSETTMR,ADMINEXPENSETTMR,FINAEXPENSETTMR,IMPAIRMENTTTMR,\
                       OPERATEINCOMETTMR,INVESTINCOMETTMR,OPTTMR,NONOPERATEPROFITTTMR,EBITTTMR,EBTTTMR,TAXTTMR,PNITTMR,KCFJCXSYJLRTTMR,\
                       NPTTMRP,FVVPALRP,IRTTMRP,IITTMFJVAJVRP,BTAATTMRP,SALESCASHINTTMR,CFOTTMR,CFITTMR,CFFTTMR,CFTTMR,CAPEXR',
                       f'ReportDate={end_date},Ispandas=1')
        if not isinstance(df_ttm, pd.DataFrame) and df_ttm.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_ttm.ErrorMsg}')
        else:
            df_list.append(df_ttm)

        if len(df_list) > 0:
            # 将每部分按列粘到一起
            df = pd.concat(df_list, axis=1)
            # 日期列修改为end_date
            df.loc[:, 'DATES'] = end_date
            df = df.reset_index().rename(columns=lambda x: EmStockFinFac.__getattribute__(EmStockFinFac, x).name)
            # 更新到db
            self._data_helper._upload_raw(df, EmStockFinFac.__table__.name)

    # 从em客户端导出数据后读取并处理
    def em_stock_daily_info_history(self, stock_daily_info_file_dir: str):
        for index, filename in enumerate(os.listdir(stock_daily_info_file_dir)):
            if not filename.endswith(".xls"):
                continue
            index += 1
            print(f'{index}: {filename}')
            stock_daily_info = pd.concat(pd.read_excel(os.path.join(stock_daily_info_file_dir, filename), sheet_name=None, na_values='--'),
                                         ignore_index=True)
            # 将列名字改为从API获取到数据时的列名字，便于后边统一处理
            stock_daily_info = stock_daily_info.drop(['简称'], axis=1).rename(columns={
                '代码': 'CODES',
                '时间': 'DATES',
                '总股本(股)': 'TOTALSHARE',
            })

            stock_daily_info = stock_daily_info.loc[~stock_daily_info['DATES'].isna()].rename(columns=lambda x: EmStockDailyInfo.__getattribute__(EmStockDailyInfo, x).name)

            # 更新到db
            self._data_helper._upload_raw(stock_daily_info, EmStockDailyInfo.__table__.name)

    def em_stock_daily_info(self, start_date: str, end_date: str, predict_year: int = 0):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_daily_info] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_daily_info] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return

            if predict_year == 0:
                predict_year = datetime.datetime.strptime(end_date, '%Y%m%d').date().year

            df_list = []
            for date in tradedates.Data:
                temp_df = c.css(stock_id_result.Codes,
                                'TOTALSHARE,HOLDFROZENAMTACCUMRATIO,PETTMDEDUCTED,PBLYRN,PSTTM,AHOLDER,ESTPEG',
                                f'EndDate={date},TradeDate={date},PREDICTYEAR={predict_year},Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock daily info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 日期列修改为end_date
                temp_df.loc[:, 'DATES'] = date
                df_list.append(temp_df)

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockDailyInfo.__getattribute__(EmStockDailyInfo, x).name)
            # 更新到db
            self._data_helper._upload_raw(df, EmStockDailyInfo.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_stock_refinancing(self, start_date: str, end_date: str):
        try:
            data = c.ctr("AdditionPlanInfo",
                         "SECURITYCODE,APPROVENOTICEDATE,PLANNOTICEDDATE,NEWPROGRESS,SUMFINA_T,ATTACHNAME,ADDPURPOSE",
                        f"DateType=2,StartDate={start_date},EndDate={end_date},Process=0")
            if data.ErrorCode != 0:
                print(f'failed to get stock refinancing info: {data.ErrorMsg}')
                return False

            infos: List[List] = []
            for v in data.Data.values():
                # if v[3] in ('董事会批准', '董事会修改', '董事会终止', '股东大会批准', '股东大会修改'):
                #    continue
                infos.append(v)
            df = pd.DataFrame(infos, columns=['SECURITYCODE', 'APPROVENOTICEDATE', 'PLANNOTICEDDATE',
                                              'NEWPROGRESS', 'SUMFINA_T', 'ATTACHNAME', 'ADDPURPOSE'])
            self._data_helper._upload_raw(df.rename(columns=lambda x: EmStockRefinancing.__getattribute__(EmStockRefinancing, x).name), EmStockRefinancing.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_stock_estimate_fac(self, start_date: str, end_date: str, predict_year: int = 0):
        try:
            # 获取区间内所有交易日
            df = RawDataApi().get_em_tradedates(start_date, end_date)
            if df is None or df.shape[0] == 0:
                print(f'[em_stock_estimate_fac] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return False

            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return False

            if predict_year == 0:
                predict_year = datetime.datetime.strptime(end_date, '%Y%m%d').date().year + 1

            df_list = []
            for date in df.TRADEDATES:
                temp_df = c.css(stock_id_result.Codes,
                                'ESTGR,ESTNI',
                                f'EndDate={date},PREDICTYEAR={predict_year},Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame):
                    if temp_df.ErrorCode != 0:
                        print(f'failed to get stock daily info: {temp_df.ErrorMsg} (date){date}')
                    return False
                # 日期列修改为end_date
                temp_df.loc[:, 'DATES'] = date
                df_list.append(temp_df)

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockEstimateFac.__getattribute__(EmStockEstimateFac, x).name)
            df = df[df.notna().all(axis=1)]
            df['predict_year'] = predict_year
            # 更新到db
            self._data_helper._upload_raw(df, EmStockEstimateFac.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index_val_history(self, start_date: str, end_date: str, index_list: List[str]):
        # 获取区间内所有交易日
        trade_day_df = RawDataApi().get_em_tradedates(start_date, end_date)
        if trade_day_df is None or trade_day_df.shape[0] == 0:
            print(f'[em_index_val_history] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
            return

        df_list: List[pd.DataFrame] = []
        # 一天一天地获取
        for date in trade_day_df.TRADEDATES:
            data = c.csd(index_list,
                        "PETTM,PBMRQ,PSTTM,PCFTTM,ROE,EPSTTM,DIVIDENDYIELD",
                        date, date,
                        "DelType=1,period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1")
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'failed to get index val info: {data.ErrorMsg} (date){date}')
                break

            _y = pd.to_datetime(date).date().year + 1
            another_data = c.css(index_list, "ESTPEG", f"TradeDate={date},PREDICTYEAR={_y},Ispandas=1")
            if not isinstance(another_data, pd.DataFrame):
                if another_data.ErrorCode != 0:
                    print(f'failed to get index val info part 2: {another_data.ErrorMsg} (date){date}')
                break
            df_list.append(data.join(another_data.drop(columns='DATES'), how='outer'))

        if df_list:
            if len(df_list) == 1:
                df = df_list[0]
            else:
                df = pd.concat(df_list)
            df = df[df.drop(columns=['DATES']).notna().any(axis=1)]
            df = df.reset_index().rename(columns={'DIVIDENDYIELD': 'DIVIDENDRATETTM'}).rename(columns=lambda x: EmIndexVal.__getattribute__(EmIndexVal, x).name)
            self._data_helper._upload_raw(df, EmIndexVal.__table__.name)

    def em_index_val(self, start_date: str, end_date: str):
        try:
            # 获取区间内所有交易日
            trade_day_df = RawDataApi().get_em_tradedates(start_date, end_date)
            if trade_day_df is None or trade_day_df.shape[0] == 0:
                print(f'[em_index_val] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return False

            # 获取basic table中index_info信息
            index_info_df = BasicDataApi().get_index_info()
            if index_info_df is None:
                print(f'[em_index_val] get index info failed')
                return False
            em_indexes = index_info_df[index_info_df.tag_method.isin(['PE百分位', 'PB百分位', 'PS百分位'])].em_id
            # 过滤掉几个特殊指数
            em_indexes = em_indexes[em_indexes.notna() & ~em_indexes.isin(['LIGBP1M.IR', 'HIBOR3M.IR', 'AU9999.SGE', 'AU0.SHF'])]
            index_list = em_indexes.to_list()
            if not index_list:
                print(f'[em_index_val] empty index list, return immediately')
                return False

            df_list: List[pd.DataFrame] = []
            # 一天一天地获取
            for date in trade_day_df.TRADEDATES:
                data = c.csd(index_list,
                           "PETTM,PBMRQ,PSTTM,PCFTTM,ROE,EPSTTM",
                           date, date,
                           "DelType=1,period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1")
                if not isinstance(data, pd.DataFrame):
                    if data.ErrorCode != 0:
                        print(f'failed to get index val info: {data.ErrorMsg} (date){date}')
                    break

                # Caution!! 股息率补历史的话需要使用csd中的DIVIDENDYIELD，这里的DIVIDENDRATETTM只能用作每日更新
                _y = pd.to_datetime(date).date().year + 1
                another_data = c.css(index_list, "DIVIDENDRATETTM,ESTPEG", f"TradeDate={date},PREDICTYEAR={_y},Ispandas=1")
                if not isinstance(another_data, pd.DataFrame):
                    if another_data.ErrorCode != 0:
                        print(f'failed to get index val info part 2: {another_data.ErrorMsg} (date){date}')
                    break
                # 这里除以100与股息率历史数据保持单位一致
                another_data['DIVIDENDRATETTM'] /= 100
                df_list.append(data.join(another_data.drop(columns='DATES'), how='outer'))

            if df_list:
                if len(df_list) == 1:
                    df = df_list[0]
                else:
                    df = pd.concat(df_list)
                df = df.reset_index().rename(columns=lambda x: EmIndexVal.__getattribute__(EmIndexVal, x).name)
                self._data_helper._upload_raw(df, EmIndexVal.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_tradedates(self, start_date, end_date, is_automatic_update=True):
        try:
            if is_automatic_update:
                # 如果是每日自动更新，我们需要去存T+1日的数据，这里保险起见多取30天的
                end_date_dt: datetime.datetime = datetime.datetime.strptime(end_date, '%Y%m%d')
                extra_end_date = (end_date_dt + datetime.timedelta(days=30)).strftime('%Y%m%d')
            else:
                extra_end_date = end_date
            # 这里确保返回结果按日期升序排序
            tradedates_result = c.tradedates(start_date, extra_end_date, "period=1,order=1,market=CNSESH")
            if tradedates_result.ErrorCode != 0:
                print(f'Failed to get tradedates: {tradedates_result.ErrorMsg}')
                return False

            tradedates = tradedates_result.Data
            if is_automatic_update:
                # 如果是每日自动更新，看一下返回的第一个日期是否是start_date，如果不是，表明今天不是交易日
                start_date_dt: datetime.datetime = pd.to_datetime(start_date, infer_datetime_format=True)
                if start_date_dt != pd.to_datetime(tradedates[0], infer_datetime_format=True):
                    print(f'today {start_date} is not a trading day')
                    return False

            df = pd.DataFrame(tradedates, columns=['TRADEDATES'])
            if is_automatic_update:
                # 如果是每日自动更新，向DB中insert T+1日的日期
                df = df.iloc[[1], :]
            self._data_helper._upload_raw(df, EmTradeDates.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_index_component(self, start_date: str, end_date: str):
        try:
            index_info = BasicDataApi().get_index_info()

            date_list = self._get_date_list(start_date, end_date)
            s_list: List[pd.Series] = []
            for row in index_info[~index_info['em_plate_id'].isna()].loc[:, ['index_id', 'em_id', 'em_plate_id']].itertuples(index=False):
                for date in date_list:
                    stock_id_result = c.sector(row.em_plate_id, date)
                    if stock_id_result.ErrorCode != 0:
                        # print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                        continue

                    stock_id_list: List[str] = []
                    for i in range(0, len(stock_id_result.Data), 2):
                        stock_id_list.append(stock_id_result.Data[i])
                    if len(stock_id_list) != 0:
                        s_list.append(pd.Series(row._asdict()).append(pd.Series({'datetime': date, 'stock_list': ','.join(stock_id_list)})))

            # 更新到db
            self._data_helper._upload_raw(pd.DataFrame(s_list), EmIndexComponent.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return

    @staticmethod
    def _lambda_func_index_component_1(x: pd.Series):
        value: List[Dict[str, str]] = json.loads(x)
        new_value = []
        for one in value:
            one = (*one.keys(), float(*one.values()))
            new_value.append(one)
        new_value.sort(key=itemgetter(1), reverse=True)
        return json.dumps(new_value)

    @staticmethod
    def _lambda_func_index_component_2(em_id_map: Dict[str, str], x: pd.Series):
        try:
            return em_id_map[x]
        except KeyError:
            return x

    @staticmethod
    def _lambda_func_index_component_3(x: pd.Series):
        value: List[Dict[str, str]] = json.loads(x)
        new_value = []
        for one in value:
            one = [*one.keys(), float(*one.values())]
            if one[0][0] in ('0', '3'):
                one[0] += '.SZ'
            elif one[0][0] in ('6'):
                one[0] += '.SH'
            new_value.append(one)
        new_value.sort(key=itemgetter(1), reverse=True)
        return json.dumps(new_value)

    @staticmethod
    def _lambda_func_index_component_4(x: pd.Series):
        if pd.isnull(x):
            return x
        return json.dumps([one + '.OF' for one in x.array])

    def cs_index_component(self, date: str):
        try:
            em_id_list = BasicDataApi().get_index_info().em_id.to_list()

            # 中证指数爬下来的部分
            columns_list = ['id', 'name', 'num', 'id_cat', 'sector', 'top10', 'related_funds']
            df_list: List[pd.DataFrame] = []
            for cat in ('industry', 'topic', 'strat', 'cap'):
                raw_df = pd.read_excel(f'./data/Index/{cat}_data.xlsx', na_values='--')
                df_list.append(raw_df[columns_list])

            df = pd.concat(df_list)
            df['id'] = df['id'].transform(lambda x: EmRawDataDownloader._lambda_func_index_component_2({em_id.split('.')[0]: em_id for em_id in em_id_list if em_id is not None}, x))
            df['id_cat'] = df['id_cat'].map({'行业': SectorType.industry.name, '主题': SectorType.topic.name, '策略': SectorType.strategy.name, '规模': SectorType.scale.name})
            df['sector'] = df['sector'].transform(EmRawDataDownloader._lambda_func_index_component_1)
            df['top10'] = df['top10'].transform(EmRawDataDownloader._lambda_func_index_component_3)
            df = df.rename(columns={'id': 'index_id'})
            # Choice中指数的全部成分及权重
            # 该表主要提供指定日期的指数成分股代码及权重等信息
            choice_data = defaultdict(list)
            for em_id in df['index_id'].to_list():
                data = c.ctr("INDEXCOMPOSITION","SECUCODE,NAME,WEIGHT",f"IndexCode={em_id},EndDate={date}")
                if data.ErrorCode != 0:
                    print(f'[cs_index_component] (em_id){em_id} (err_msg){data.ErrorMsg}')
                    continue
                for constituent in data.Data.values():
                    choice_data[em_id].append((constituent[0], round(constituent[2] * 100, 2)))
                choice_data[em_id].sort(key=itemgetter(1), reverse=True)
                choice_data[em_id] = json.dumps(choice_data[em_id])

            df['datetime'] = date
            df = df.set_index('index_id').assign(all_constitu=pd.Series(choice_data))
            # 更新到db
            self._data_helper._upload_raw(df.drop(columns='name').reset_index(), CSIndexComponent.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return

    def em_fund_list(self, start_date: str, end_date: str):
        try:
            # 这两个成分只能取到最新的，所以直接用end_date即可，没必要遍历
            # 获取所有基金列表（不包含未成立、已到期）
            fund_list_result = c.sector('507013', end_date)
            if fund_list_result.ErrorCode != 0:
                print(f'failed to get live fund list: {fund_list_result.ErrorMsg}')
                return
            # 获取已摘牌基金列表
            delisted_fund_list_result = c.sector('507022', end_date)
            if delisted_fund_list_result.ErrorCode != 0:
                print(f'failed to get delisted fund list: {delisted_fund_list_result.ErrorMsg}')
                return

            df = pd.DataFrame({'datetime': [end_date], 'all_live_fund_list': [','.join(fund_list_result.Codes)],
                               'delisted_fund_list': [','.join(delisted_fund_list_result.Codes)]})
            # 更新到db
            self._data_helper._upload_raw(df, EmFundList.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return

    def _em_get_all_fund_list(self, date: str):
        # 获取全部基金
        all_fund = c.sector('507013', date)
        if all_fund.ErrorCode != 0:
            print(f'failed to get all fund list: {all_fund.ErrorMsg}')
            return

        # 获取全部基金列表（含未成立、已到期）
        all_fund_extra_full = c.sector('507010', date)
        if all_fund_extra_full.ErrorCode != 0:
            print(f'failed to get all fund list: {all_fund_extra_full.ErrorMsg}')
            return

        # 获取全部基金列表（含未成立）
        all_fund_extra = c.sector('507009', date)
        if all_fund_extra.ErrorCode != 0:
            print(f'failed to get all fund list: {all_fund_extra.ErrorMsg}')
            return

        # 不应该保存未成立基金的数据，这里我们需要(全部 + (含未成立、已到期 - 含未成立))
        all_fund_list = set(all_fund.Codes)
        all_fund_extra_full_list = set(all_fund_extra_full.Codes)
        all_fund_extra_list = set(all_fund_extra.Codes)
        all_fund_list.update(all_fund_extra_full_list.difference(all_fund_extra_list))
        return all_fund_list

    def em_fund_info_history(self, date: str):
        fund_list = self._em_get_all_fund_list(date)
        if fund_list is None:
            return

        # 更新wind_fund_info中没有的数据
        wind_df = RawDataApi().get_wind_fund_info()
        self.em_fund_info(fund_list.difference(wind_df.wind_id))

    def em_fund_info(self, fund_id_list: Set[str]):
        if not fund_id_list:
            return True
        try:
            # 过滤掉非.OF结尾的基金
            fund_id_list = [one for one in fund_id_list if one.split('.')[1] == 'OF']
            if not fund_id_list:
                return True
            data = c.css(','.join(fund_id_list), 'NAME,FNAME,FUNDTYPE,FOUNDDATE,MATURITYDATE,FIRSTINVESTTYPE,SECONDINVESTTYPE,TRADECUR,STRUCFUNDORNOT,GFCODEM,PREDFUNDMANAGER,MGRCOMP', 'Ispandas=1')
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'failed to get fund info: {data.ErrorMsg}')
                return False

            data['FUNDTYPE'] = data['FUNDTYPE'].map(lambda x: True if x == '契约型开放式' else False)
            data['STRUCFUNDORNOT'] = data['STRUCFUNDORNOT'].map(lambda x: True if x == '是' else False)
            df = data.drop(columns='DATES').reset_index().rename(columns=lambda x: EmFundInfo.__getattribute__(EmFundInfo, x).name)
            # 更新到db，先删再添，确保数据不会冲突，且有一个更新的机会
            # TODO: 最好原子提交下边两个操作
            RawDataApi().delete_em_fund_info(df.em_id.to_list())
            self._data_helper._upload_raw(df, EmFundInfo.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_fund_fee_history(self, date: str):
        fund_list = self._em_get_all_fund_list(date)
        if fund_list is None:
            return
        self.em_fund_fee(all_fund_list)

    @staticmethod
    def _lambda_fund_fee_1(x):
        if x.CODEFEEMODE.array[0] == CodeFeeMode.none:
            fee_type = 1
        else:
            fee_type = x.CODEFEEMODE.array[0].value
        data = c.css(','.join(x.index.to_list()), 'MANAGFEERATIO,CUSTODIANFEERATIO,PURCHFEERATIO,REDEMFEERATIO,SERVICEFEERATIO,SUBSCRFEERATIO', f'FeeType={fee_type},Ispandas=1')
        if not isinstance(data, pd.DataFrame):
            if data.ErrorCode != 0:
                print(f'failed to get fund code fee: {data.ErrorMsg}')
            return
        return data

    def em_fund_fee(self, fund_id_list: Set[str]):
        if not fund_id_list:
            return True
        try:
            data = c.css(','.join(fund_id_list), 'CODEFEEMODE', 'Ispandas=1')
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'failed to get fund code fee mode: {data.ErrorMsg}')
                return False
            # 这里没有判断css的返回值
            data['CODEFEEMODE'] = data['CODEFEEMODE'].map({'前端': CodeFeeMode.a_type, '后端': CodeFeeMode.b_type, None: CodeFeeMode.none})
            data = data.groupby(by='CODEFEEMODE', sort=False).apply(EmRawDataDownloader._lambda_fund_fee_1).reset_index()
            data['CODEFEEMODE'] = data['CODEFEEMODE'].transform(lambda x: x.name)
            df = data.drop(columns='DATES').rename(columns=lambda x: EmFundFee.__getattribute__(EmFundFee, x).name)
            # 更新到db，先删再添，确保数据不会冲突，且有一个更新的机会
            # TODO: 最好原子提交下边两个操作
            RawDataApi().delete_em_fund_fee(df.em_id.to_list())
            self._data_helper._upload_raw(df, EmFundFee.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_fund_benchmark_history(self, start_date: str, end_date: str):
        # 获取所有基金列表（含未成立、已到期）
        fund_list_result = c.sector('507010', end_date)
        if fund_list_result.ErrorCode != 0:
            print(f'failed to get all fund list: {fund_list_result.ErrorMsg}')
            return

        df = RawDataApi().get_em_tradedates(start_date, end_date)
        for row in df.itertuples(index=False):
            self.em_fund_benchmark(row.TRADEDATES, fund_list_result.Codes)
            print(f'[em_fund_benchmark_history] {row.TRADEDATES} done')

    def em_fund_benchmark(self, end_date: str, fund_list):
        try:
            data = c.css(fund_list, 'BENCHMARK', f'EndDate={end_date},Ispandas=1')
            if not isinstance(data, pd.DataFrame):
                if data.ErrorCode != 0:
                    print(f'failed to get fund benchmark: {data.ErrorMsg}')
                return False
            data = data[data.notna().all(axis=1)]
            if data.empty:
                return True

            df = data.reset_index().rename(columns=lambda x: EmFundBenchmark.__getattribute__(EmFundBenchmark, x).name)
            df['datetime'] = end_date
            # 获取当前的benchmark
            now_df = RawDataApi().get_em_fund_benchmark(end_date)
            if now_df is not None:
                # 过滤出来新增基金benchmark以及旧基金的benchmark发生变化的情况
                df = df.merge(now_df.drop(columns=['_update_time']), how='left', on=['em_id', 'benchmark'], indicator=True, validate='one_to_one')
                df = df[df._merge == 'left_only'].drop(columns=['_merge', 'datetime_y']).rename(columns={'datetime_x': 'datetime'})
            # 更新到db
            self._data_helper._upload_raw(df, EmFundBenchmark.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_fund_rate(self, end_date: str):
        try:
            fund_list = c.sector('507013', end_date)
            fund_list = fund_list.Codes
            df=c.css(fund_list,"MKTCOMPRE3YRAT,MORNSTAR3YRAT,MORNSTAR5YRAT,MERCHANTS3YRAT,TXCOMRATRAT,SHSTOCKSTAR3YCOMRAT,SHSTOCKSTAR3YSTOCKRAT,SHSTOCKSTAR3YTIMERAT,SHSTOCKSTAR3YSHARPERAT,SHSTOCKSTAR5YCOMRATRAT,SHSTOCKSTAR5YSTOCKRAT,SHSTOCKSTAR5YTIMERAT,SHSTOCKSTAR5YSHARPRAT,JAJXCOMRAT,JAJXEARNINGPOWERRAT,JAJXACHIEVESTABILITYRAT,JAJXANTIRISKRAT,JAJXSTOCKSELECTIONRAT,JAJXTIMESELECTIONRAT,JAJXBENCHMARKTRACKINGRAT,JAJXEXCESSEARNINGSRAT,JAJXTOTALFEERAT",
                f"EndDate={end_date},Ispandas=1")
            df['DATES'] = datetime.datetime.strptime(end_date,'%Y%m%d').date()
            df = df.reset_index()
            self._data_helper._upload_raw(df, EMFundRate.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def em_fund_asset(self):
        all_fund = c.sector('507013', '20200719').Codes
        dt_list = ['20191231','20200331','20200630']
        res = []
        for dt in dt_list:
            data=c.css(all_fund,"PRTSTOCKTONAV,PRTBONDTONAV,PRTCASHTONAV,PRTOTHERTONAV",f"ReportDate={dt},Ispandas=1")
            data['DATES'] = dt
            res.append(data)
        df = pd.concat(res).reset_index()
        self._data_helper._upload_raw(df, 'em_fund_asset')

    def em_fund_industry(self):
        all_fund = c.sector('507013', '20200719').Codes
        dt_list = ['20191231','20200331','20200630']
        res = []
        rank_list = [1,2,3]
        for dt in dt_list:
            _res = [] 
            for rank in rank_list:
                data=c.css(all_fund,"PRTKEYINDNAME,PRTKEYINDTOASSET",f"ReportDate={dt},Rank={rank},IndustryCode=A,NewOldType=1,Ispandas=1")
                data = data.rename(columns={'PRTKEYINDNAME':f'rank{rank}_indname','PRTKEYINDTOASSET':f'rank{rank}_indweight'})
                data = data.drop('DATES',axis=1)
                _res.append(data)
            df = pd.concat(_res,axis=1)
            df['DATES'] = dt
            res.append(df)
        df = pd.concat(res).reset_index()
        self._data_helper._upload_raw(df, 'em_fund_industry')

    def em_fund_stock(self):
        all_fund = c.sector('507013', '20200719').Codes
        dt_list = ['20191231','20200331','20200630']
        res = []
        rank_list = [i for i in range(1,11,1)]
        for dt in dt_list:
            _res = [] 
            for rank in rank_list:
                data=c.css(all_fund,"PRTKEYSTOCKNAME,PRTKEYSTOCKCODE,PRTKEYSTOCKVALUE,PRTKEYSTOCKTONAV",f"ReportDate={dt},Rank={rank},Ispandas=1")
                data = data.rename(columns={'PRTKEYSTOCKNAME': f'rank{rank}_stock',
                                            'PRTKEYSTOCKCODE': f'rank{rank}_stock_code',
                                            'PRTKEYSTOCKVALUE': f'rank{rank}_stockval',
                                            'PRTKEYSTOCKTONAV': f'rank{rank}_stockweight',
                                            })
                data = data.drop('DATES',axis=1)
                _res.append(data)
                print(f'[em_fund_stock] {dt} rank {rank} done')
            df = pd.concat(_res,axis=1)
            df['DATES'] = dt
            res.append(df)
        df = pd.concat(res).reset_index()
        self._data_helper._upload_raw(df, 'em_fund_stock')

    def em_fund_bond(self):
        all_fund = c.sector('507013', '20200719').Codes
        fund_list = all_fund
        dt_list = ['20191231','20200331','20200630']
        res = []
        rank_list = [i for i in range(1,11,1)]
        for dt in dt_list:
            _res = [] 
            for rank in rank_list:
                data=c.css(fund_list,"PRTKEYBONDNAME,PRTKEYBONDCODE,PRTKEYBONDVALUE,PRTKEYBONDTONAV",f"ReportDate={dt},Rank={rank},Ispandas=1")
                data = data.rename(columns={'PRTKEYBONDNAME':f'rank{rank}_bond','PRTKEYBONDCODE':f'rank{rank}_bond_code','PRTKEYBONDVALUE':f'rank{rank}_bondval','PRTKEYBONDTONAV':f'rank{rank}_bondweight'})
                data = data.drop('DATES',axis=1)
                _res.append(data)
            df = pd.concat(_res,axis=1)
            df['DATES'] = dt
            res.append(df)
        df = pd.concat(res).reset_index()
        em._data_helper._upload_raw(df, 'em_fund_bond')

    def download_all(self, start_date, end_date):
        failed_tasks = []
        # If em_tradedates fails, there is no trading day between start_date and end_date
        # Stop and return
        if not self.em_tradedates(start_date, end_date):
            failed_tasks.append('em_tradedates')
            return failed_tasks

        # could only update the end_date's data
        if not self.em_fund_status(end_date):
            failed_tasks.append('em_fund_status')

        if not self.em_fund_nav(start_date, end_date):
            failed_tasks.append('em_fund_nav')

        index_info_df = BasicDataApi().get_index_info()
        if index_info_df is None:
            failed_tasks.append('retrieve data for em_index_price')
        else:
            index_info_df = index_info_df[index_info_df.price_source == IndexPriceSource.default]
            index_ids = set(index_info_df.em_id.to_list())
            index_ids.discard(None)
            if not self.em_index_price(index_ids, start_date, end_date):
                failed_tasks.append('em_index_price')

        if not self.em_stock_info(end_date):
            failed_tasks.append('em_stock_info')

        if not self.em_stock_price(start_date, end_date):
            failed_tasks.append('em_stock_price')

        if not self.em_stock_post_price(start_date, end_date):
            failed_tasks.append('em_stock_post_price')

        if not self.em_index_val(start_date, end_date):
            failed_tasks.append('em_index_val')

        if not self.em_stock_daily_info(start_date, end_date):
            failed_tasks.append('em_stock_daily_info')

        if not self.em_index_component(start_date, end_date):
            failed_tasks.append('em_index_component')

        if not self.em_fund_list(start_date, end_date):
            failed_tasks.append('em_fund_list')
        else:
            new_fund_list, _ = RawDataHelper.get_new_and_delisted_fund_list(end_date)
            if new_fund_list is None:
                failed_tasks.append('get new_fund_list')
            else:
                if not self.em_fund_info(new_fund_list):
                    failed_tasks.append('em_fund_info')
                if not self.em_fund_fee(new_fund_list):
                    failed_tasks.append('em_fund_fee')
                # if not self.em_fund_benchmark(end_date, new_fund_list):
                #     failed_task.append('em_fund_benchmark')

        if not self.em_macroeconomic_price(start_date, end_date):
            failed_tasks.append('em_macroeconomic_price')

        if not self.em_fund_rate(end_date):
            failed_tasks.append('em_fund_rate')

        return failed_tasks


if __name__ == '__main__':
    em = EmRawDataDownloader(RawDataHelper())
    # em.em_tradedates('20000101', '20091231')
    # em.em_index_val('20200428', '20200428')
    # em.em_fund_nav('20200415', '20200427')
    # em.em_index_price_history('H00140.SH', '20050101', '20200419')
    # em.em_fund_scale_history()
    # em.em_stock_fin_fac('20161231')
    # em.em_stock_info('20200424')
    # em.em_stock_price('20041231', '20041231')
    # em.em_stock_post_price('20041231', '20041231')
    # em.em_stock_daily_info('20041231', '20041231')
    # em.em_stock_price_history('./')
    # em.em_stock_post_price_history('./')
    # em.em_stock_daily_info_history('./')
    # em.em_stock_refinancing('20170101', '20181231')
    # em.add_columns_to_fin_fac('20190930')
    # em.em_fund_status('20200513')
    # em.em_index_component('20200528', '20200529')
    # em.em_fund_list('20200702', '20200702')
    # em.em_fund_info_history('20200707')
    # em.em_fund_benchmark_history('20070101', '20071231')
    # em.em_index_info(['399972.SZ'])
    # em.em_fund_info({'165807.SZ'})
    # em.em_stock_estimate_fac('20200101', '20200709', 2020)
    # em.cs_index_component('20200713')
    # em.em_macroeconomic_price('20200714', '20200714')
    # em.em_fund_fee_history('20200715')




from typing import Tuple

import math
import pandas as pd
import pickle as pkl
from ..wrapper.mysql import DerivedDatabaseConnector
from ..view.derived_models import *


class DerivedDataApi(object):
    def get_fund_indicator(self, fund_list):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                        FundIndicator.fund_id,
                        FundIndicator.datetime,
                        FundIndicator.alpha,
                        FundIndicator.beta,
                        FundIndicator.fee_rate,
                        FundIndicator.track_err,
                    ).filter(
                        FundIndicator.fund_id.in_(fund_list),
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicator.__tablename__}')

    def get_fund_indicator_by_date(self, fund_list, date):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundIndicator
                ).filter(
                    FundIndicator.datetime == date,
                    FundIndicator.fund_id.in_(fund_list),
                )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicator.__tablename__}')

    def get_fund_indicator_monthly(self, start_date, end_date, fund_list, columns: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundIndicatorMonthly.fund_id,
                    FundIndicatorMonthly.datetime,
                )
                if columns:
                    query = query.add_columns(*columns)
                query = query.filter(
                    FundIndicatorMonthly.fund_id.in_(fund_list),
                    FundIndicatorMonthly.datetime >= start_date,
                    FundIndicatorMonthly.datetime <= end_date
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicatorMonthly.__tablename__}')

    def get_fund_score(self):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                        FundScore,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScore.__tablename__}')

    def get_fund_score_extended_by_id(self, fund_id):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundScoreExtended,
                ).filter(
                    FundScoreExtended.fund_id == fund_id,
                )

                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScoreExtended.__tablename__}')
                return pd.DataFrame([])

    def get_fund_score_extended(self, start_date: str, end_date: str, fund_list: Tuple[str] = (), columns: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                if columns:
                    query = quant_session.query(
                        FundScoreExtended.fund_id,
                        FundScoreExtended.datetime,
                    )
                    query = query.add_columns(*columns)
                else:
                    query = quant_session.query(
                        FundScoreExtended,
                    )
                if fund_list:
                    query = query.filter(
                        FundScoreExtended.fund_id.in_(fund_list)
                    )
                query = query.filter(
                    FundScoreExtended.datetime >= start_date,
                    FundScoreExtended.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScoreExtended.__tablename__}')

    def get_fund_score_extended_for_ranks(self, start_date: str, end_date: str, fund_list: Tuple[str] = ()):
        return self.get_fund_score_extended(start_date, end_date, fund_list=fund_list, columns=['wind_class_1', 'return_score', 'robust_score', 'timing_score', 'return_rank', 'robust_rank', 'timing_rank'])

    def get_index_valuation_develop(self, index_ids, start_date, end_date):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuation
                ).filter(
                    IndexValuation.index_id.in_(index_ids),
                    IndexValuation.datetime >= start_date,
                    IndexValuation.datetime <= end_date,
                ).order_by(IndexValuation.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuation.__tablename__}')
                return pd.DataFrame([])

    def get_index_valuation_develop_without_date(self, index_ids):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuation
                ).filter(
                    IndexValuation.index_id.in_(index_ids),
                ).order_by(IndexValuation.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuation.__tablename__}')
                return pd.DataFrame([])

    def get_asset_allocation_info(self, version:int=2):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    AssetAllocationInfo
                ).filter(
                    AssetAllocationInfo.version == version
                ).order_by(AssetAllocationInfo.allocation_id)
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {AssetAllocationInfo.__tablename__}')
                return pd.DataFrame([])

    def get_style_factor_return(self, start_date: str, end_date: str, index_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    StyleAnalysisFactorReturn
                )
                if index_list:
                    query = query.filter(StyleAnalysisFactorReturn.universe_index.in_(index_list))
                query = query.filter(
                    StyleAnalysisFactorReturn.datetime >= start_date,
                    StyleAnalysisFactorReturn.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_style_factor_return {e} from {StyleAnalysisFactorReturn.__tablename__}')
                return

    def get_allocation_distribution(self, version:str=1):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    AllocationDistribution
                ).filter(
                    AllocationDistribution.version == version
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get_style_factor_return {e} from {AllocationDistribution.__tablename__}')
                return





import datetime
import traceback
import pandas as pd
from cassandra.cqlengine.query import DoesNotExist
from cassandra.cqlengine.query import BatchQuery

from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.cas.fund_estimated_nav import FundEstimatedNav


class FundEstimatedNavProcessor:

    def rq_stock_code_to_em(self, rq_stock_code):
        if not rq_stock_code:
            return None
        em_stock_code = rq_stock_code.replace("XSHG", "SH")
        em_stock_code = rq_stock_code.replace("XSHE", "SZ")
        return em_stock_code

    def load_fund_stock(self):
        fund_stock_df = BasicDataApi().get_fund_hold_stock()
        # print(f'fund_stock_df\n {fund_stock_df}')
        fund_stock_info_list = [(fund_stock_df.columns.get_loc(f'rank{i}_stock_code'), 
            fund_stock_df.columns.get_loc(f'rank{i}_stockweight')) for i in range(1,11,1)]
        fund_stock = {}
        for row in fund_stock_df.itertuples(index=False):
            stock_info = []
            for fund_stock_info in fund_stock_info_list:
                if not row[fund_stock_info[0]]:
                    break
                stock_info.append([row[fund_stock_info[0]], row[fund_stock_info[1]]])
            if not stock_info:
                continue
            fund_stock[row.fund_id] = stock_info
        return fund_stock

    def load_fund_nav(self, dt):
        fund_nav_df = BasicDataApi().get_fund_nav(dt=dt)
        # print(f'fund_nav_df\n {fund_nav_df}')
        fund_nav = {}
        for row in fund_nav_df.itertuples(index=False):
            fund_nav[row.fund_id] = [row.unit_net_value, row.acc_net_value, row.adjusted_net_value]
        return fund_nav

    def load_stock_minute(self, stock_minute_df=None):
        if stock_minute_df is None:
            stock_minute_df = RawDataApi().get_rq_stock_minute()
        # print(f'stock_minute_df\n {stock_minute_df}')
        if stock_minute_df is None or stock_minute_df.empty:
            return None

        curr_datetime = stock_minute_df.iloc[0].datetime
        
        stock_minute = {}
        for row in stock_minute_df.itertuples(index=False):
            em_stock_code = self.rq_stock_code_to_em(row.order_book_id)
            if not em_stock_code:
                continue
            stock_minute[em_stock_code] = row.close
        return curr_datetime, stock_minute

    def load_stock_price(self, dt):
        em_stock_price_df = RawDataApi().get_em_stock_price(dt, dt, columns=['close'])
        # print(f'em_stock_price_df\n {em_stock_price_df}')
        stock_price = {}
        for row in em_stock_price_df.itertuples(index=False):
            stock_price[row.stock_id] = row.close
        return stock_price

    def delta_estimated(self, stock_weight, stock_price_last_trading_day, stock_price_now):
        return stock_weight / 100.0 * (stock_price_now / stock_price_last_trading_day - 1)

    def calc_fund_estimated_nav(self, stock_minute_df=None):
        try:
            stock_minute_parse_result = self.load_stock_minute(stock_minute_df)
            if not stock_minute_parse_result:
                return False
            curr_datetime, stock_minute = stock_minute_parse_result
            fund_stock = self.load_fund_stock()

            last_trading_day = BasicDataApi().get_last_trading_day()
            print(f'last_trading_day: {last_trading_day}')
            if not last_trading_day:
                return False
            fund_nav_last_trading_day = self.load_fund_nav(last_trading_day)
            stock_price_last_trading_day = self.load_stock_price(last_trading_day)

            results = []
            for fund_id, stock_list in fund_stock.items():
                if fund_id not in fund_nav_last_trading_day:
                    continue

                ratio = 1.0
                for stock in stock_list:
                    if stock[0] not in stock_price_last_trading_day or stock[0] not in stock_minute:
                        continue
                    curr_stock_price_last_trading_day = stock_price_last_trading_day[stock[0]]
                    curr_stock_price_now = stock_minute[stock[0]]

                    # TODO: this stock component weight is not evaluated at last_trading_day!!!
                    # TODO: re-calc stock component weight for last_trading_day to get a more accurate estimation!
                    ratio += self.delta_estimated(
                        stock[1], curr_stock_price_last_trading_day, curr_stock_price_now)

                unit_net_value_estimated = fund_nav_last_trading_day[fund_id][0] * ratio
                acc_net_value_estimated = fund_nav_last_trading_day[fund_id][1] * ratio
                adjusted_net_value_estimated = fund_nav_last_trading_day[fund_id][2] * ratio
                
                results.append([fund_id, unit_net_value_estimated, acc_net_value_estimated, adjusted_net_value_estimated, curr_datetime])

                if len(results) >= 200:
                    # Write Cassandra DB
                    with BatchQuery() as batch:
                        for res in results:
                            FundEstimatedNav.objects(fund_id=res[0]).batch(batch).update(
                                unit_net_values__append = [res[1]],
                                acc_net_values__append = [res[2]],
                                adjusted_net_values__append = [res[3]],
                                timestamps__append = [res[4]],
                            )
                    results = []
            
            if results:
                # Write Cassandra DB
                with BatchQuery() as batch:
                    for res in results:
                        FundEstimatedNav.objects(fund_id=res[0]).batch(batch).update(
                            unit_net_values__append = [res[1]],
                            acc_net_values__append = [res[2]],
                            adjusted_net_values__append = [res[3]],
                            timestamps__append = [res[4]],
                        )
            # result_df = pd.DataFrame(results, columns = ['fund_id', 'unit_net_value', 'acc_net_value', 'adjusted_net_value', 'datetime']) 
            # print(result_df)
            # self._data_helper._upload_basic(result_df, FundEstimatedNav.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False


if __name__ == '__main__':
    bdp = BasicDataPart3(BasicDataHelper())




from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import AssetTrade, FundTrade, FundPosition, FundWeight
from ...data.struct import AssetTradeParam, FundTradeParam
from ...data.struct import AssetTradeCache, FundWeightItem
from ...data.constant import TradeTrigger
from . import Helper
from .asset_helper import FAHelper
import numpy as np
import math
import datetime

class AssetTrader(Helper):

    TRADE_DAY_INTERVAL = 10

    def __init__(self, asset_param: AssetTradeParam=None):
        self.asset_param = asset_param or AssetTradeParam()
        self.last_trade_date = None

    def trade_interval_allow(self, dt):
        if self.last_trade_date is None or self.last_trade_date + datetime.timedelta(days=self.TRADE_DAY_INTERVAL) < dt:
            return True
        else:
            return False

    def calc_asset_trade(self, dt,
                               cur_position: AssetPosition,
                               cur_price: AssetPrice,
                               target_allocation: AssetWeight):
        cur_mv = AssetValue(prices=cur_price, positions=cur_position)
        tot_mv = cur_mv.sum()

        trades = []
        launch_trade = False
        for index_id, target_weight in target_allocation.__dict__.items():
            if index_id != 'cash':
                target_amt = tot_mv * target_weight
                p = cur_price.__dict__[index_id]
                cur_amt = cur_mv.__dict__[index_id]
                if abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinCountAmtDiff:
                    amount = abs(target_amt - cur_amt)
                    is_buy = target_amt > cur_amt
                    if is_buy:
                        trades.append(AssetTrade(
                            index_id=index_id,
                            mark_price=p,
                            amount=amount,
                            is_buy=is_buy,
                            submit_date=dt
                        ))
                    else:
                        trades.append(AssetTrade(
                            index_id=index_id,
                            mark_price=p,
                            volume=amount/p,
                            is_buy=is_buy,
                            submit_date=dt
                        ))
                launch_trade = launch_trade or abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinActionAmtDiff

        if not launch_trade:
            return cur_position, None
        else:
            trades.sort(key=lambda x: x.is_buy)
            new_position = cur_position.copy()
            for trd in trades:
                new_position.update(trd)
            #如果一个交易单过小，不达到交易下限，会被忽略，导致买单和卖单交易额相加不相等
            buy_trades_amt = sum([i.amount for i in trades if i.is_buy is True])
            sell_trades_amt = sum([i.mark_price * i.volume for i in trades if i.is_buy is False])
            buy_adjusted = sell_trades_amt / buy_trades_amt if (sell_trades_amt > 0 and buy_trades_amt > 0) else 1
            for trd in trades:
                if trd.is_buy == True:
                    trd.amount *= buy_adjusted
            # buy_trades_amt = sum([i.amount for i in trades if i.is_buy is True])
            # sell_trades_amt = sum([i.mark_price * i.volume for i in trades if i.is_buy is False]) 
            # print(f' compare trades amt buy {buy_trades_amt} and sell {sell_trades_amt} diff {buy_trades_amt-sell_trades_amt }')
            return new_position, trades

    def finalize_trade(self, dt, trades: list,
                            t1_price: AssetPrice,
                            bt_position: AssetPosition):
        pendings = []
        traded_list = []
        if trades is None or len(trades) == 0:
            return pendings, traded_list
        for trd in trades:
            trd.trade_price = t1_price.__dict__[trd.index_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.trade_date = dt
            if not trd.is_buy:
                if not(bt_position.__dict__[trd.index_id] - trd.volume > -1e-8):
                    #print(f'trade volume exceeds, adjusted to pos (index_id){trd.index_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){bt_position.__dict__[trd.index_id]}')
                    trd.volume = bt_position.__dict__[trd.index_id]
            trd.amount = trd.volume * trd.trade_price    
            if self.asset_param.EnableCommission:
                if not trd.is_buy:
                    trd.commission = trd.amount * self.asset_param.RedeemDiscount * self.asset_param.AssetRedeemRate[trd.index_id]
                    trd.amount -= trd.commission

        _trade_list_buy = [i  for i in trades if i.is_buy is True]
        _trade_list_sel = [i  for i in trades if i.is_buy is False]
        # 买单按照amount下单， 不会出现计划资金和实际资金不符合情况
        # 卖单按照volume下单， 会出现计算资金和实际资金不符合情况
        # 保证卖单和买单实际成交的金额一致，就不会出现现金比例波动的情况，现金就可以设置为0，去掉现金资产比例 
        plan_sel_amount = 0 if len(_trade_list_sel) < 1 else sum([i.amount for i in _trade_list_sel])
        plan_buy_amount = 0 if len(_trade_list_buy) < 1 else sum([i.amount for i in _trade_list_buy])
        if plan_sel_amount == 0 or plan_buy_amount == 0: 
            buy_amount_adjusted_rate = 1
        else:
            buy_amount_adjusted_rate = plan_sel_amount / plan_buy_amount
        for trd in trades:
            if trd.is_buy:
                trd.amount *= buy_amount_adjusted_rate
            if self.asset_param.EnableCommission:
                if trd.is_buy:
                    trd.volume = trd.amount / trd.trade_price / (1 + self.asset_param.PurchaseDiscount * self.asset_param.AssetPurchaseRate[trd.index_id])
                    trd.commission = trd.volume * trd.trade_price * self.asset_param.PurchaseDiscount * self.asset_param.AssetPurchaseRate[trd.index_id]
            else:
                trd.commission = 0
            bt_position.update(trd)
            traded_list.append(trd)
        
        self.last_trade_date = dt
        return pendings, traded_list

class FundTrader(AssetTrader):

    SMALL_POS_FLOAT = 1e-6
    BIG_INT = 1e10

    def __init__(self, asset_param: AssetTradeParam=None, fund_param: FundTradeParam=None):
        AssetTrader.__init__(self, asset_param=asset_param)
        self.fund_param = fund_param or FundTradeParam()

    def set_helper(self, fa_helper: FAHelper):
        self.fa_helper = fa_helper
        self.last_trade_date = None

    def has_expired_fund(self, cur_fund_position:FundPosition, _prep_fund_score:dict):
         # 如果持仓基金 没有分数， 返回True
        pos_fund_set = {fund_id for fund_id, fund_pos_i in cur_fund_position.funds.items() if fund_pos_i.volume > 0}
        score_set = set()
        for index_id, pos_i in _prep_fund_score.items():
            score_set.update(pos_i.keys())
        return bool(pos_fund_set.difference(score_set))

    # to be deprecated
    def calc_fund_trade(self, dt, fund_weight: FundWeight, cur_fund_position: FundPosition,
                            cur_fund_nav: dict,
                            fund_purchase_fees: dict,
                            fund_redeem_fees: dict) -> list:
        new_fund_position = cur_fund_position.copy()
        fund_trades = []
        # return trade list
        fund_tot_mv, cur_fund_wgts = cur_fund_position.calc_mv_n_w(fund_navs=cur_fund_nav)
        all_funds = {}
        # prepare fund candidates and its index_id
        for fund_id, fund_wgt_item in fund_weight.funds.items():
            all_funds[fund_id] = fund_wgt_item.index_id
        for fund_id, fund_pos_item in cur_fund_position.funds.items():
            all_funds[fund_id] = fund_pos_item.index_id
        # calc trade
        for fund_id, index_id in all_funds.items():
            target_fund_amt = fund_weight.get_wgt(fund_id) * fund_tot_mv
            cur_fund_volume = cur_fund_position.get_volume(fund_id) or 0
            p = cur_fund_nav[fund_id]
            cur_fund_amt = cur_fund_volume * p

            if abs(target_fund_amt - cur_fund_amt) > fund_tot_mv * self.fund_param.MinCountAmtDiff or (target_fund_amt == 0 and cur_fund_amt > 0):
                # TODO: commision and 如果是清某一只基金的逻辑，清空可以执行
                is_buy = target_fund_amt > cur_fund_amt
                if is_buy:
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=index_id,
                        mark_price=p,
                        amount=abs(target_fund_amt - cur_fund_amt),
                        is_buy=is_buy,
                        submit_date=dt
                    )
                else:
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=index_id,
                        mark_price=p,
                        volume=abs(target_fund_amt - cur_fund_amt)/p,
                        is_buy=is_buy,
                        submit_date=dt
                    )
                fund_trades.append(_trade)
                #print(f'(fund){fund_id} (p){p} (amt0){cur_fund_amt} (amt1){target_fund_amt} (idx){index_id} (amt){abs(target_fund_amt - cur_fund_amt)} (direc) {target_fund_amt > cur_fund_amt} ')

        fund_trades.sort(key=lambda x: x.is_buy)
        for _trade in fund_trades:
            new_fund_position.update(_trade)
        return new_fund_position, fund_trades

    def finalize_trade(self, dt, trades: list,
                            t1_price: AssetPrice,
                            bt_position: AssetPosition,
                            cur_fund_position: FundPosition,
                            cur_fund_nav: dict,
                            cur_fund_unit_nav: dict,
                            fund_purchase_fees: dict,
                            fund_redeem_fees: dict,
                            disproved_set: set):

        if trades is None or len(trades) == 0:
            return [], []

        pendings = []
        traded_list = []
        # TODO: if some trades needs more time
        # 先成交卖单， 再根据卖单成交金额 等额成交买单
        for trd in trades:
            trd.trade_price = cur_fund_nav[trd.fund_id]
            trd.fund_unit_nav = cur_fund_unit_nav[trd.fund_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.amount = trd.amount if trd.amount else (trd.volume * trd.trade_price)
            trd.fund_unit_volume = trd.amount / trd.fund_unit_nav
            trd.trade_date = dt
            if not trd.is_buy:
                cur_vol = cur_fund_position.get_volume(trd.fund_id)
                if not((cur_vol or 0) - trd.volume > -1e-8):
                    #print(f'trade volume exceeds, adjusted to pos (fund_id){trd.fund_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){cur_vol}')
                    assert cur_vol is not None, 'sell fund with no current position!'
                    trd.volume = cur_vol
                    trd.amount = trd.volume * trd.trade_price
            
            if self.fund_param.EnableCommission:
                if not trd.is_buy:
                    redeem_fee = fund_redeem_fees[trd.fund_id] * self.fund_param.RedeemDiscount
                    if np.isnan(redeem_fee):
                        #print(f'fund_id {trd.fund_id} redeem fee data not avaiable')
                        redeem_fee = 0
                    trd.commission = trd.amount * redeem_fee
                    trd.amount -= trd.commission
            else:
                trd.commission = 0
            if not trd.is_buy:
                trade_status = cur_fund_position.update(trd)
                if trade_status:
                    traded_list.append(trd)
        _trade_list_buy = [i  for i in trades if i.is_buy is True]
        _trade_list_sel = [i  for i in traded_list if i.is_buy is False]
        plan_sel_amount = 0 if len(_trade_list_sel) < 1 else sum([i.amount for i in _trade_list_sel])
        plan_buy_amount = 0 if len(_trade_list_buy) < 1 else sum([i.amount for i in _trade_list_buy])
        if plan_sel_amount == 0 or plan_buy_amount == 0: 
            buy_amount_adjusted_rate = 1
        else:
            buy_amount_adjusted_rate = plan_sel_amount / plan_buy_amount
        
        for trd in trades:
            if not trd.is_buy:
                continue
            trd.amount *= buy_amount_adjusted_rate
            if self.fund_param.EnableCommission:
                purchase_fee = fund_purchase_fees[trd.fund_id] * self.fund_param.PurchaseDiscount 
                if np.isnan(purchase_fee):
                    #print(f'fund_id {trd.fund_id} purchase fee data not avaiable')
                    purchase_fee = 0
                trd.volume = trd.amount / trd.trade_price / (1 + purchase_fee)
                trd.commission = trd.volume * trd.trade_price * purchase_fee
            else:
                trd.commission = 0            
                #print(f'trade is not permitted : {trd}')
            trade_status = cur_fund_position.update(trd)
            if trade_status:
                traded_list.append(trd)
            else:
                pass
                #print(f'trade failed alert : {trd}')
        _trade_list_buy = [i  for i in traded_list if i.is_buy is True]
        _trade_list_sel = [i  for i in traded_list if i.is_buy is False]
        # plan_sel_amount = 0 if len(_trade_list_sel) < 1 else sum([i.amount for i in _trade_list_sel])
        # plan_buy_amount = 0 if len(_trade_list_buy) < 1 else sum([i.amount for i in _trade_list_buy])
        # print(f'trade date {dt} sell amount {plan_sel_amount} buy amount {plan_buy_amount} diff {plan_buy_amount-plan_sel_amount}')

        # get cur asset weight from fund weight
        cur_mv, cur_fund_weight = cur_fund_position.calc_mv_n_w(cur_fund_nav)
        fund_index_dic = {}
        for fund_id, fund_pos_item in cur_fund_position.funds.items():
            fund_index_dic[fund_id] = fund_pos_item.index_id
        asset_wgt = { _ : 0 for _ in set(fund_index_dic.values())}
        for fund_id, wgt_i in cur_fund_weight.items():
            index_id = fund_index_dic[fund_id]
            asset_wgt[index_id] += wgt_i
        # set index position:
        for index_id in asset_wgt:
            asset_p = t1_price.__dict__[index_id]
            amount = cur_mv * asset_wgt[index_id]
            bt_position.__dict__[index_id] = amount / asset_p
        
        self.last_trade_date = dt
        # fund cash -> asset cash
        bt_position.cash = cur_fund_position.cash
        return pendings, traded_list

    def judge_trade(self, dt, 
            index_fund_cache: list, # to change
            tar_fund_weight: FundWeight,
            tar_asset_weight: AssetWeight,
            cur_fund_position: FundPosition,
            cur_asset_position: AssetPosition,
            cur_fund_nav: dict,
            cur_fund_score: dict,
            cur_asset_price: AssetPrice) -> bool:

        index_rebalance_max_diff = 0
        fund_selection_min_score = 1
        fund_rebalance_min_score = 1
        fund_end = 0
        _index_rebalance_max_diff_name = ''
        _fund_selection_min_name = ''
        _fund_rebalance_min_name = ''
        fund_rank_list = []
        cur_asset_mv = AssetValue(prices=cur_asset_price, positions=cur_asset_position)
        cur_asset_weight = cur_asset_mv.get_weight()
        for index_id, index_tar_wgt in tar_asset_weight.__dict__.items():
            if index_id == 'cash':
                continue
            c = AssetTradeCache(index_id=index_id)
            c.index_tar_wgt = index_tar_wgt
            c.index_cur_wgt = cur_asset_weight.__dict__[index_id]
            c.index_diff = c.index_tar_wgt - c.index_cur_wgt
            if abs(c.index_diff) > index_rebalance_max_diff:
                index_rebalance_max_diff = abs(c.index_diff)
                _index_rebalance_max_diff_name = index_id

            # calc funds in each index
            c.fund_cur_wgts = cur_fund_position.calc_mv_n_w(cur_fund_nav, index_id)[1]
            c.cur_fund_ids = set(c.fund_cur_wgts.keys())
            c.fund_scores = sorted(cur_fund_score.get(index_id, {}).items(), key=lambda item: item[1], reverse=True)
            c.fund_ranks = {info[0]: rank + 1 for rank, info in enumerate(c.fund_scores)}
            c.proper_fund_num = min(self.fa_helper.get_max_fund_num(index_id), len(c.fund_scores)) if index_tar_wgt > self.SMALL_POS_FLOAT else 0
            # 维度1：现有基金的选取优化情况：（评估整体现有基金池的好坏）
            c.fund_judge_ranking_score = 0
            wgt_func = lambda _rank: 1.0 / math.pow(_rank, 1/3)
            _rank_list = []
            for f in c.cur_fund_ids:
                rank_i = c.fund_ranks.get(f, self.BIG_INT)
                if rank_i == self.BIG_INT:
                    fund_end = 1
                _rank_list.append((f, rank_i))
                c.fund_judge_ranking_score += wgt_func(rank_i)

            c.fund_judge_ranking_best = 0
            for i in range(0, c.proper_fund_num):
                c.fund_judge_ranking_best += wgt_func(i + 1)
            c.fund_judge_ranking = (c.fund_judge_ranking_score + self.SMALL_POS_FLOAT) / (c.fund_judge_ranking_best + self.SMALL_POS_FLOAT)
            
            # 维度2：现有基金的比例平均化情况：（评估现有基金比例的合理性）
            if len(c.cur_fund_ids) != c.proper_fund_num:
                c.fund_judge_diverse = 0
            else:
                c.fund_judge_diverse = 1
                for fund_id in c.cur_fund_ids:
                    old_wgt = c.fund_cur_wgts.get(fund_id, 0)
                    c.fund_judge_diverse *= old_wgt * c.proper_fund_num

            # to start or others, cur_fund_ids is few
            if len(c.cur_fund_ids) < c.proper_fund_num:
                for fund_id, _score in c.fund_scores:
                    if fund_id not in c.cur_fund_ids:
                        c.cur_fund_ids.add(fund_id)
                        if len(c.cur_fund_ids) == c.proper_fund_num:
                            break
            assert len(c.cur_fund_ids) >= c.proper_fund_num, 'cur fund id candidate should be no less than proper_fund_num'
            
            if index_tar_wgt > self.fund_param.DiffJudgeAssetWgtRequirement:
                if fund_selection_min_score > c.fund_judge_ranking:
                    fund_selection_min_score = c.fund_judge_ranking
                    _fund_selection_min_name = index_id
                    fund_rank_list = _rank_list

                if fund_rebalance_min_score > c.fund_judge_diverse:
                    fund_rebalance_min_score = c.fund_judge_diverse
                    _fund_rebalance_min_name = index_id
            index_fund_cache.append(c)

        # self.fund_param.DiffJudgeLambda = 0
        trigger = 0
        trigger += TradeTrigger.IndexRebalance if index_rebalance_max_diff > self.fund_param.JudgeIndexDiff else 0
        trigger += TradeTrigger.FundSelection if fund_selection_min_score < self.fund_param.JudgeFundSelection else 0
        trigger += TradeTrigger.FundRebalance if fund_rebalance_min_score < self.fund_param.JudgeFundRebalance else 0
        trigger += TradeTrigger.FundEnd if fund_end == 1 else 0

        trigger_reason = TradeTrigger.trigger_log(trigger,
                                                  index_rebalance_max_diff,
                                                  _index_rebalance_max_diff_name,
                                                  self.fund_param.JudgeIndexDiff,
                                                  fund_selection_min_score,
                                                  _fund_selection_min_name,
                                                  self.fund_param.JudgeFundSelection,
                                                  fund_rank_list,
                                                  fund_rebalance_min_score,
                                                  _fund_rebalance_min_name,
                                                  self.fund_param.JudgeFundRebalance)
        
        
        if trigger > 0:
            if self.trade_interval_allow(dt):
                #print(f'judge allow {dt} (trigger){TradeTrigger.parse(trigger)} (Fs){fund_selection_min_score}[{_fund_selection_min_name}]#{self.fund_param.JudgeFundSelection} (Fr){fund_rebalance_min_score}/[{_fund_rebalance_min_name}]#{self.fund_param.JudgeFundRebalance} (Ir){index_rebalance_max_diff}/[{_index_rebalance_max_diff_name}]#{self.fund_param.JudgeIndexDiff}')
                pass
            else:
                trigger = 0
                #print(f'judge dismiss {dt} (trigger){TradeTrigger.parse(trigger)} (Fs){fund_selection_min_score}[{_fund_selection_min_name}]#{self.fund_param.JudgeFundSelection} (Fr){fund_rebalance_min_score}/[{_fund_rebalance_min_name}]#{self.fund_param.JudgeFundRebalance} (Ir){index_rebalance_max_diff}/[{_index_rebalance_max_diff_name}]#{self.fund_param.JudgeIndexDiff}')
        return trigger, trigger_reason

    def calc_trade(self, dt,
            tar_fund_weight: FundWeight,
            tar_asset_weight: AssetWeight,
            cur_fund_position: FundPosition,
            cur_asset_position: AssetPosition,
            cur_fund_nav: dict,
            cur_fund_score: dict,
            cur_asset_price: AssetPrice) -> list:
        '''
        v_asset_position, asset_trade_list = self.calc_asset_trade(dt, cur_asset_position, cur_asset_price, tar_asset_weight)
        if not asset_trade_list:
            return v_asset_position, asset_trade_list
        else:
            return self.calc_fund_trade(dt, tar_fund_weight, cur_fund_position, cur_fund_nav, fund_purchase_fees, fund_redeem_fees)
        '''
        index_fund_cache = []

        trigger, trigger_reason = self.judge_trade(dt, index_fund_cache, 
                                    tar_fund_weight, tar_asset_weight,
                                    cur_fund_position, cur_asset_position, cur_fund_nav, 
                                    cur_fund_score, cur_asset_price)
        
        if trigger == 0:
            return cur_fund_position, [], {}

        fund_trades = self.get_trades(dt, index_fund_cache, 
                                    tar_fund_weight, tar_asset_weight,
                                    cur_fund_position, cur_asset_position, cur_fund_nav, 
                                    cur_fund_score, cur_asset_price)


        fund_trades.sort(key=lambda x: x.is_buy)
        new_fund_position = cur_fund_position.copy()

        for _trade in fund_trades:
            _trade.trigger = trigger
            new_fund_position.update(_trade)
        return new_fund_position, fund_trades, trigger_reason


    def get_trades(self, dt,
            index_fund_cache, 
            tar_fund_weight: FundWeight,
            tar_asset_weight: AssetWeight,
            cur_fund_position: FundPosition,
            cur_asset_position: AssetPosition,
            cur_fund_nav: dict,
            cur_fund_score: dict,
            cur_asset_price: AssetPrice):
        
        fund_tot_mv, cur_fund_wgts = cur_fund_position.calc_mv_n_w(fund_navs=cur_fund_nav)
        
        fund_trades = []

        for c in index_fund_cache:
            index_id = c.index_id
            tar_fund_list = tar_fund_weight.get_funds(index_id)
            cur_fund_list = cur_fund_position.get_funds(index_id)
            for fund_id in set(tar_fund_list).union(cur_fund_list):
                target_fund_amt = tar_fund_weight.get_wgt(fund_id) * fund_tot_mv
                cur_fund_volume = cur_fund_position.get_volume(fund_id) or 0
                p = cur_fund_nav[fund_id]
                cur_fund_amt = cur_fund_volume * p

                if abs(target_fund_amt - cur_fund_amt) > fund_tot_mv * self.fund_param.MinCountAmtDiff \
                    or (target_fund_amt == 0 and cur_fund_amt > 0):
                    # 如果是清某一只基金的逻辑，清空可以执行
                    # commision 放到 finalize_trade 部分执行
                    is_buy = target_fund_amt > cur_fund_amt
                    if is_buy:
                        _trade = FundTrade(
                            fund_id=fund_id,
                            index_id=index_id,
                            mark_price=p,
                            amount=abs(target_fund_amt - cur_fund_amt),
                            is_buy=is_buy,
                            submit_date=dt
                        )
                    else:
                        is_to_clean_up = target_fund_amt == 0 and cur_fund_amt > 0
                        _trade = FundTrade(
                            fund_id=fund_id,
                            index_id=index_id,
                            mark_price=p,
                            # 如果清仓 在 calc_trade时就把volume 改成全部仓位， 防止小量仓位残留
                            volume= cur_fund_volume if is_to_clean_up else abs(target_fund_amt - cur_fund_amt)/p,
                            is_buy=is_buy,
                            submit_date=dt,
                            is_to_cleanup=is_to_clean_up
                        )
                    fund_trades.append(_trade)
                    # print(f'(fund){fund_id} (d){"buy" if is_buy else "sell"} (r){c.fund_ranks.get(fund_id, -1)} (p){p} (amt){target_fund_amt - cur_fund_amt} (tar){target_fund_amt} (ind){index_id}')
        
        #如果一个交易单过小，不达到交易下限，会被忽略，导致买单和卖单交易额相加不相等
        buy_trades_amt = sum([i.amount for i in fund_trades if i.is_buy is True])
        sell_trades_amt = sum([i.mark_price * i.volume for i in fund_trades if i.is_buy is False]) 
        buy_adjusted = sell_trades_amt / buy_trades_amt if (sell_trades_amt > 0 and buy_trades_amt > 0) else 1
        for trd in fund_trades:
            if trd.is_buy == True:
                trd.amount *= buy_adjusted
        
        ## 检查买卖量是否一致
        # buy_trades_amt = sum([i.amount for i in fund_trades if i.is_buy is True])
        # sell_trades_amt = sum([i.mark_price * i.volume for i in fund_trades if i.is_buy is False]) 
        # print(f'  compare trades dt {dt} amt buy {buy_trades_amt} and sell {sell_trades_amt} diff {buy_trades_amt-sell_trades_amt }')
        return fund_trades



from .cas_connection import CassandraConnector
from ...view.cas.fund_estimated_nav import FundEstimatedNav
from cassandra.cqlengine.management import sync_table, drop_table


def sync_cas_table():
    sync_table(FundEstimatedNav)

def drop_cas_table():
    drop_table(FundEstimatedNav)


if __name__ == '__main__':
    CassandraConnector().get_conn()
    sync_cas_table()
    # drop_cas_table()
